#!/usr/bin/python3

from collections.abc import Sequence
from typing import Final

import eth_hash.auto
from eth_typing import ABIComponent, ABIFunction, HexStr

keccak: Final = eth_hash.auto.keccak

_cached_int_bounds: Final[dict[str, tuple[int, int]]] = {}


def get_int_bounds(type_str: str) -> tuple[int, int]:
    """Returns the lower and upper bound for an integer type."""
    try:
        return _cached_int_bounds[type_str]
    except KeyError:
        # validate input
        size = int(type_str.strip("uint") or 256)
        if size < 8 or size > 256 or size % 8:
            raise ValueError(f"Invalid type: {type_str}")

        # compute lower and upper bound
        if type_str.startswith("u"):
            lower = 0
            upper = 2**size - 1
        else:
            lower = -(2 ** (size - 1))
            upper = 2 ** (size - 1) - 1

        # cache result and return
        _cached_int_bounds[type_str] = lower, upper
        return lower, upper


def get_type_strings(
    abi_params: Sequence[ABIComponent],
    substitutions: dict[str, str] | None = None,
) -> list[str]:
    """Converts a list of parameters from an ABI into a list of type strings."""
    types_list = []
    if substitutions is None:
        substitutions = {}

    for i in abi_params:
        type_str = i["type"]
        if type_str.startswith("tuple"):
            params = get_type_strings(i["components"], substitutions)
            array_size = type_str[5:]
            types_list.append(f"({','.join(params)}){array_size}")
        else:
            for orig, sub in substitutions.items():
                if type_str.startswith(orig):
                    type_str = type_str.replace(orig, sub)
            types_list.append(type_str)

    return types_list


def build_function_signature(abi: ABIFunction) -> str:
    types_list = get_type_strings(abi["inputs"])
    return f"{abi['name']}({','.join(types_list)})"


def build_function_selector(abi: ABIFunction) -> HexStr:
    sig = build_function_signature(abi)
    return f"0x{keccak(sig.encode()).hex()[:8]}"  # type: ignore [return-value]



#!/usr/bin/python3

import shutil
import sys
from typing import Any, Final, cast

import yaml

from brownie._c_constants import Path
from brownie._config import CONFIG, _get_data_folder
from brownie.utils import color, notify
from brownie.utils._color import bright_black, bright_magenta, green
from brownie.utils.docopt import docopt

__doc__ = """Usage: brownie networks <command> [<arguments> ...] [options]

Commands:
  list [verbose=false]             List existing networks
  add <env> <id> [key=value, ...]  Add a new network
  modify <id> [key=value, ...]     Modify field(s) for an existing network
  import <path> <replace=False>    Import network settings
  export <path>                    Export network settings
  delete <id>                      Delete an existing network
  update_provider <name> <url>     Adds or modifies a new network provider
  delete_provider <name>           Removes a network provider
  set_provider <name>              Sets a provider from the list of providers
  list_providers                   List available providers

Options:
  --help -h                        Display this message

Settings related to local development chains and live environments.

Each network has a unique id. To connect to a specific network when running tests
or launching the console, use the commandline flag `--network [id]`.

To add a network you must specify an environment and id, as well as required fields.
For example, to add a network "mainnet" to the "Ethereum" environment:

  brownie networks add Ethereum mainnet host=https://mainnet.infura.io/ chainid=1

Use `brownie networks list true` to see a detailed view of available networks
as well as possible data fields when declaring new networks."""


DEV_REQUIRED: Final = ("id", "host", "cmd", "cmd_settings")
PROD_REQUIRED: Final = ("id", "host", "chainid")
OPTIONAL: Final = ("name", "explorer", "timeout", "multicall2", "provider")

DEV_CMD_SETTINGS: Final = frozenset(
    [
        "port",
        "gas_limit",
        "accounts",
        "evm_version",
        "fork",
        "disable_cache",
        "mnemonic",
        "account_keys_path",
        "block_time",
        "default_balance",
        "time",
        "network_id",
        "chain_id",
        "unlimited_contract_size",
    ]
)


def main() -> None:
    args = docopt(__doc__)
    try:
        fn = getattr(sys.modules[__name__], f"_{args['<command>']}")
    except AttributeError:
        print("Invalid command. Try brownie networks --help")
        return
    try:
        fn(*args["<arguments>"])
    except TypeError:
        print(f"Invalid arguments for command '{args['<command>']}'. Try brownie networks --help")
        return


def _list(verbose: bool | str = False) -> None:
    if isinstance(verbose, str):
        try:
            verbose = cast(bool, eval(verbose.capitalize()))
        except (NameError, SyntaxError) as e:
            print("Please pass 'True' or 'False'.")
            raise e

    with _get_data_folder().joinpath("network-config.yaml").open() as fp:
        networks = yaml.safe_load(fp)

    print("The following networks are declared:")

    for chain in networks["live"]:
        print(f"\n{chain['name']}")
        chain_networks = chain["networks"]
        for value in chain_networks:
            is_last = value == chain_networks[-1]
            if verbose:
                _print_verbose_network_description(value, is_last)
            else:
                _print_simple_network_description(value, is_last)

    print("\nDevelopment")
    dev_networks = networks["development"]
    for value in dev_networks:
        is_last = value == dev_networks[-1]
        if verbose:
            settings = value.pop("cmd_settings")
            _print_verbose_network_description(value, is_last)
            _print_verbose_network_description(settings, is_last, 2)
        else:
            _print_simple_network_description(value, is_last)


def _add(env: str, id_: str, *args: str) -> None:
    if id_ in CONFIG.networks:
        raise ValueError(f"Network '{bright_magenta}{id_}{color}' already exists")

    args_dict = _parse_args(args)

    if "name" not in args_dict:
        args_dict["name"] = id_

    with _get_data_folder().joinpath("network-config.yaml").open() as fp:
        networks = yaml.safe_load(fp)
    if env.lower() == "development":
        try:
            new = {
                "name": args_dict.pop("name"),
                "id": id_,
                "cmd": args_dict.pop("cmd"),
                "host": args_dict.pop("host"),
            }
        except KeyError as exc:
            raise ValueError(f"Missing field: {exc.args[0]}")
        if "timeout" in args_dict:
            new["timeout"] = args_dict.pop("timeout")
        new["cmd_settings"] = args_dict
        _validate_network(new, DEV_REQUIRED)
        networks["development"].append(new)
    else:
        target = next(
            (i["networks"] for i in networks["live"] if i["name"].lower() == env.lower()), None
        )
        if target is None:
            networks["live"].append({"name": env, "networks": []})
            target = networks["live"][-1]["networks"]
        new = {"id": id_, **args_dict}
        _validate_network(new, PROD_REQUIRED)
        target.append(new)
    with _get_data_folder().joinpath("network-config.yaml").open("w") as fp:
        yaml.dump(networks, fp)

    notify("SUCCESS", f"A new network '{bright_magenta}{new['name']}{color}' has been added")
    _print_verbose_network_description(new, True)


def _modify(id_: str, *args: str) -> None:
    if id_ not in CONFIG.networks:
        raise ValueError(f"Network '{bright_magenta}{id_}{color}' does not exist")

    args_dict = _parse_args(args)

    with _get_data_folder().joinpath("network-config.yaml").open() as fp:
        networks = yaml.safe_load(fp)

    is_dev = "cmd" in CONFIG.networks[id_]
    if is_dev:
        target = next(i for i in networks["development"] if i["id"] == id_)
    else:
        target = next(x for i in networks["live"] for x in i["networks"] if x["id"] == id_)
    for key, value in args_dict.items():
        t = target
        if key in DEV_CMD_SETTINGS and is_dev:
            t = target["cmd_settings"]
        if value is None:
            del t[key]
        else:
            t[key] = value
    if is_dev:
        _validate_network(target, DEV_REQUIRED)
    else:
        _validate_network(target, PROD_REQUIRED)

    if "name" not in target:
        target["name"] = id_

    with _get_data_folder().joinpath("network-config.yaml").open("w") as fp:
        yaml.dump(networks, fp)

    notify("SUCCESS", f"Network '{bright_magenta}{target['name']}{color}' has been modified")
    _print_verbose_network_description(target, True)


def _delete(id_: str) -> None:
    if id_ not in CONFIG.networks:
        raise ValueError(f"Network '{bright_magenta}{id_}{color}' does not exist")

    with _get_data_folder().joinpath("network-config.yaml").open() as fp:
        networks = yaml.safe_load(fp)

    if "cmd" in CONFIG.networks[id_]:
        networks["development"] = [i for i in networks["development"] if i["id"] != id_]
    else:
        target = next(i for i in networks["live"] for x in i["networks"] if x["id"] == id_)
        target["networks"] = [i for i in target["networks"] if i["id"] != id_]
        networks["live"] = [i for i in networks["live"] if i["networks"]]

    with _get_data_folder().joinpath("network-config.yaml").open("w") as fp:
        yaml.dump(networks, fp)

    notify("SUCCESS", f"Network '{bright_magenta}{id_}{color}' has been deleted")


def _import(path_str: str, replace: bool = False) -> None:
    if isinstance(replace, str):
        replace = eval(replace.capitalize())

    path = Path(path_str)
    with path.open() as fp:
        new_networks = yaml.safe_load(fp)

    with _get_data_folder().joinpath("network-config.yaml").open() as fp:
        old_networks = yaml.safe_load(fp)

    for value in new_networks.get("development", []):
        id_ = value["id"]
        if id_ in CONFIG.networks:
            if "cmd" not in CONFIG.networks[id_]:
                raise ValueError(
                    f"Import file contains development network with id '{id_}',"
                    " but this is already an existing live network."
                )
            if not replace:
                raise ValueError(f"Cannot overwrite existing network {id_}")
            old_networks["development"] = [i for i in old_networks["development"] if i["id"] != id_]
        _validate_network(value, DEV_REQUIRED)
        old_networks["development"].append(value)

    for chain in new_networks.get("live", []):
        for value in chain["networks"]:
            prod = next((i for i in old_networks["live"] if i["name"] == chain["name"]), None)
            if prod is None:
                prod = {"name": chain["name"], "networks": []}
                old_networks["live"].append(prod)
            id_ = value["id"]
            if id_ in CONFIG.networks:
                if not replace:
                    raise ValueError(f"Cannot overwrite existing network {id_}")
                existing = next((i for i in prod["networks"] if i["id"] == id_), None)
                if existing is None:
                    raise ValueError(
                        f"Import file contains live network with id '{id_}',"
                        " but this is already an existing network on a different environment."
                    )
                prod["networks"].remove(existing)
            _validate_network(value, PROD_REQUIRED)
            prod["networks"].append(value)

    with _get_data_folder().joinpath("network-config.yaml").open("w") as fp:
        yaml.dump(old_networks, fp)

    notify("SUCCESS", f"Network settings imported from '{bright_magenta}{path}{color}'")


def _export(path_str: str) -> None:
    path = Path(path_str)
    if path.exists():
        if path.is_dir():
            path = path.joinpath("network-config.yaml")
        else:
            raise FileExistsError(f"{path} already exists")
    if not path.suffix:
        path = path.with_suffix(".yaml")
    shutil.copy(_get_data_folder().joinpath("network-config.yaml"), path)

    notify("SUCCESS", f"Network settings exported as '{bright_magenta}{path}{color}'")


def _update_provider(name: str, url: str) -> None:
    with _get_data_folder().joinpath("providers-config.yaml").open() as fp:
        providers = yaml.safe_load(fp)

    providers[name] = {"host": url}

    with _get_data_folder().joinpath("providers-config.yaml").open("w") as fp:
        yaml.dump(providers, fp)

    notify("SUCCESS", f"Provider '{bright_magenta}{name}{color}' has been updated")


def _delete_provider(name: str) -> None:
    with _get_data_folder().joinpath("providers-config.yaml").open() as fp:
        providers = yaml.safe_load(fp)

    if name not in providers.keys():
        raise ValueError(f"Provider '{bright_magenta}{name}{color}' does not exist")

    del providers[name]

    with _get_data_folder().joinpath("providers-config.yaml").open("w") as fp:
        yaml.dump(providers, fp)

    notify("SUCCESS", f"Provider '{bright_magenta}{name}{color}' has been deleted")


def _set_provider(name: str) -> None:
    with _get_data_folder().joinpath("providers-config.yaml").open() as fp:
        providers = yaml.safe_load(fp)

    if name not in providers.keys():
        raise ValueError(f"Provider '{bright_magenta}{name}{color}' does not exist")

    with _get_data_folder().joinpath("network-config.yaml").open() as fp:
        networks = yaml.safe_load(fp)

    for blockchain in networks["live"]:
        for network in blockchain["networks"]:
            if "provider" in network.keys() and network["provider"]:
                new_network_name = network["name"].replace(
                    network["provider"].capitalize(), name.capitalize()
                )

                new_host = providers[name]["host"].format(network["id"])
                _modify(
                    network["id"],
                    f"name={new_network_name}",
                    f"provider={name}",
                    f"host={new_host}",
                )


def _list_providers(verbose: bool | str = False):
    if isinstance(verbose, str):
        try:
            verbose = cast(bool, eval(verbose.capitalize()))
        except (NameError, SyntaxError) as e:
            print("Please pass 'True' or 'False'.")
            raise e

    with _get_data_folder().joinpath("providers-config.yaml").open() as fp:
        providers: dict[str, Any] = yaml.safe_load(fp)

    print("The following providers are declared:")
    if verbose:
        _print_verbose_providers_description(providers)
    else:
        _print_simple_providers_description(providers)


def _parse_args(args: tuple[str, ...]) -> dict[str, Any]:
    try:
        parsed: dict[str, Any] = dict(i.split("=", maxsplit=1) for i in args)
    except ValueError:
        raise ValueError("Arguments must be given as key=value") from None

    for key, value in parsed.items():
        if value.isdigit():
            parsed[key] = int(value)
        elif value.lower() in ("true", "false", "none"):
            parsed[key] = eval(value.capitalize())

    return parsed


def _print_verbose_providers_description(providers: dict[str, Any]) -> None:
    u = "\u251c"
    for provider in providers:
        print(f"{bright_black}  {u}\u2500{color}provider: {provider}:")
        print(f"{bright_black}  {u}\u2500{color}   host: {providers[provider]}:")


def _print_simple_providers_description(providers: dict[str, Any]) -> None:
    u = "\u251c"
    print(f"{bright_black}  {u}\u2500{color}{providers.keys()}:")


def _print_simple_network_description(network_dict: dict[str, Any], is_last: bool) -> None:
    u = "\u2514" if is_last else "\u251c"
    print(
        f"{bright_black}  {u}\u2500{color}{network_dict['name']}:"
        f" {green}{network_dict['id']}{color}"
    )


def _print_verbose_network_description(
    network_dict: dict[str, Any], is_last: bool, indent: int = 0
) -> None:
    if is_last:
        u = "\u2514"
        v = " "
    else:
        u = "\u251c"
        v = "\u2502"
    if "name" in network_dict:
        print(f"{bright_black}  {u}\u2500{color}{network_dict.pop('name')}")

    obj_keys = sorted(network_dict)
    if "id" in obj_keys:
        obj_keys.remove("id")
        obj_keys.insert(0, "id")

    for key in obj_keys:
        value = network_dict[key]
        u = "\u2514" if key == obj_keys[-1] else "\u251c"

        if indent:
            u = (" " * indent) + u
        c = green if key == "id" else ""
        print(f"{bright_black}  {v} {u}\u2500{color}{key}: {c}{value}{color}")


def _validate_network(network: dict[str, Any], required: tuple[str, ...]) -> None:
    if missing := [i for i in required if i not in network]:
        raise ValueError(f"Network is missing required field(s): {', '.join(missing)}")

    if unknown := [i for i in network if i not in required + OPTIONAL]:
        raise ValueError(f"Unknown field(s): {', '.join(unknown)}")

    if "cmd_settings" in network:
        if unknown := [i for i in network["cmd_settings"] if i not in DEV_CMD_SETTINGS]:
            raise ValueError(f"Unknown field(s): {', '.join(unknown)}")



#!/usr/bin/python3

import logging
from typing import Any, Deque, Final

import semantic_version
import solcast
import solcx
from eth_typing import ABIElement, HexStr
from requests.exceptions import ConnectionError
from solcast.nodes import NodeBase, is_inside_offset

from brownie._c_constants import Version, deque, sha1
from brownie._config import EVM_EQUIVALENTS
from brownie.exceptions import CompilerError, IncompatibleSolcVersion  # noqa
from brownie.project.compiler.utils import VersionList, VersionSpec, _get_alias, expand_source_map
from brownie.typing import (
    BranchMap,
    BytecodeJson,
    Count,
    DeployedBytecodeJson,
    InputJsonSolc,
    IntegerString,
    Offset,
    PcList,
    PCMap,
    ProgramCounter,
    SolidityBuildJson,
    Source,
    StatementMap,
)

from . import sources

solcx_logger: Final = logging.getLogger("solcx")
solcx_logger.setLevel(10)
sh: Final = logging.StreamHandler()
sh.setLevel(10)
sh.setFormatter(logging.Formatter("%(message)s"))
solcx_logger.addHandler(sh)

AVAILABLE_SOLC_VERSIONS: VersionList | None = None

EVM_VERSION_MAPPING: Final = [
    ("prague", Version("0.8.30")),
    ("cancun", Version("0.8.25")),
    ("shanghai", Version("0.8.20")),
    ("paris", Version("0.8.18")),
    ("london", Version("0.8.7")),
    ("berlin", Version("0.8.5")),
    ("istanbul", Version("0.5.13")),
    ("petersburg", Version("0.5.5")),
    ("byzantium", Version("0.4.0")),
]

PcMap = dict[Count, ProgramCounter]
StatementNodes = dict[str, set[Offset]]
BranchNodes = dict[str, set[NodeBase]]

_BINOPS_PARAMS: Final = {"nodeType": "BinaryOperation", "typeDescriptions.typeString": "bool"}


def get_version() -> semantic_version.Version:
    return solcx.get_solc_version(with_commit_hash=True)


def compile_from_input_json(
    input_json: InputJsonSolc,
    silent: bool = True,
    allow_paths: str | None = None,
) -> dict[str, Any]:
    """
    Compiles contracts from a standard input json.

    Args:
        input_json: solc input json
        silent: verbose reporting
        allow_paths: compiler allowed filesystem import path

    Returns: standard compiler output json
    """

    settings = input_json["settings"]
    settings.setdefault("evmVersion", None)
    if settings["evmVersion"] in EVM_EQUIVALENTS:
        settings["evmVersion"] = EVM_EQUIVALENTS[settings["evmVersion"]]

    if not silent:
        print(f"Compiling contracts...\n  Solc version: {str(solcx.get_solc_version())}")

        optimizer = settings["optimizer"]
        opt = f"Enabled  Runs: {optimizer['runs']}" if optimizer["enabled"] else "Disabled"
        print(f"  Optimizer: {opt}")

        if settings["evmVersion"]:
            print(f"  EVM Version: {settings['evmVersion'].capitalize()}")

    try:
        return solcx.compile_standard(input_json, allow_paths=allow_paths)
    except solcx.exceptions.SolcError as e:
        raise CompilerError(e, "solc")


def set_solc_version(version: VersionSpec) -> str:
    """Sets the solc version. If not available it will be installed."""
    if not isinstance(version, Version):
        version = Version(version.lstrip("v"))
    if version < Version("0.4.22"):
        raise IncompatibleSolcVersion("Brownie only supports Solidity versions >=0.4.22")
    try:
        solcx.set_solc_version(version, silent=True)
    except solcx.exceptions.SolcNotInstalled:
        if version not in _get_solc_version_list()[0]:
            raise IncompatibleSolcVersion(
                f"Cannot install Solidity v{version} on this OS. You may be able to "
                f"manually compile from source with `solcx.compile_solc('{version}')`"
            )
        install_solc(version)
        solcx.set_solc_version(version, silent=True)
    return str(solcx.get_solc_version())


def install_solc(*versions: VersionSpec) -> None:
    """Installs solc versions."""
    for version in versions:
        solcx.install_solc(version, show_progress=False)


def get_abi(contract_source: str, allow_paths: str | None = None) -> dict[str, list[ABIElement]]:
    """
    Given a contract source, returns a dict of {name: abi}

    This function is deprecated in favor of `brownie.project.compiler.get_abi`
    """
    version = find_best_solc_version({"<stdin>": contract_source})
    set_solc_version(version)
    compiled: dict[str, dict] = solcx.compile_source(
        contract_source, allow_empty=True, allow_paths=allow_paths, output_values=["abi"]
    )
    return {k.rsplit(":")[-1]: v["abi"] for k, v in compiled.items()}


def find_solc_versions(
    contract_sources: dict[str, str],
    install_needed: bool = False,
    install_latest: bool = False,
    silent: bool = True,
) -> dict[str, list[str]]:
    """
    Analyzes contract pragmas and determines which solc version(s) to use.

    Args:
        contract_sources: a dictionary in the form of {'path': "source code"}
        install_needed: if True, will install when no installed version matches
                        the contract pragma
        install_latest: if True, will install when a newer version is available
                        than the installed one
        silent: set to False to enable verbose reporting

    Returns: dictionary of {'version': ['path', 'path', ..]}
    """

    available_versions, installed_versions = _get_solc_version_list()

    pragma_specs: dict = {}
    to_install = set()
    new_versions = set()

    for path, source in contract_sources.items():
        pragma_spec = sources.get_pragma_spec(source, path)
        pragma_specs[path] = pragma_spec
        version = pragma_spec.select(installed_versions)

        if not version and not install_needed and not install_latest:
            raise IncompatibleSolcVersion(
                f"No installed solc version matching '{pragma_spec}' in '{path}'"
            )

        # if no installed version of solc matches the pragma, find the latest available version
        latest = pragma_spec.select(available_versions)

        if not version and not latest:
            raise IncompatibleSolcVersion(
                f"No installable solc version matching '{pragma_spec}' in '{path}'"
            )

        if not version or (install_latest and latest > version):
            to_install.add(latest)
        elif latest and latest > version:
            new_versions.add(str(version))

    # install new versions if needed
    if to_install:
        install_solc(*to_install)
        installed_versions = solcx.get_installed_solc_versions()
    elif new_versions and not silent:
        print(
            f"New compatible solc version{'s' if len(new_versions) > 1 else ''}"
            f" available: {', '.join(new_versions)}"
        )

    # organize source paths by latest available solc version
    compiler_versions: dict[str, list[str]] = {}
    for path, spec in pragma_specs.items():
        version = spec.select(installed_versions)
        compiler_versions.setdefault(str(version), []).append(path)

    return compiler_versions


def find_best_solc_version(
    contract_sources: dict[str, str],
    install_needed: bool = False,
    install_latest: bool = False,
    silent: bool = True,
) -> str:
    """
    Analyzes contract pragmas and finds the best version compatible with all sources.

    Args:
        contract_sources: a dictionary in the form of {'path': "source code"}
        install_needed: if True, will install when no installed version matches
                        the contract pragma
        install_latest: if True, will install when a newer version is available
                        than the installed one
        silent: set to False to enable verbose reporting

    Returns: version string
    """
    available_versions: VersionList
    installed_versions: VersionList

    available_versions, installed_versions = _get_solc_version_list()

    for path, source in contract_sources.items():
        pragma_spec = sources.get_pragma_spec(source, path)
        installed_versions = [i for i in installed_versions if i in pragma_spec]
        available_versions = [i for i in available_versions if i in pragma_spec]

    if not available_versions:
        raise IncompatibleSolcVersion("No installable solc version compatible across all sources")

    if not installed_versions and not install_needed and not install_latest:
        raise IncompatibleSolcVersion("No installed solc version compatible across all sources")

    if max(available_versions) > max(installed_versions, default=Version("0.0.0")):
        if install_latest or (install_needed and not installed_versions):
            install_solc(max(available_versions))
            return str(max(available_versions))
        if not silent:
            print(f"New compatible solc version available: {max(available_versions)}")

    return str(max(installed_versions))


def _get_solc_version_list() -> tuple[VersionList, VersionList]:
    global AVAILABLE_SOLC_VERSIONS
    installed_versions: VersionList = solcx.get_installed_solc_versions()
    if AVAILABLE_SOLC_VERSIONS is None:
        try:
            AVAILABLE_SOLC_VERSIONS = solcx.get_installable_solc_versions()
        except ConnectionError:
            if not installed_versions:
                raise ConnectionError("Solc not installed and cannot connect to GitHub")
            AVAILABLE_SOLC_VERSIONS = installed_versions
    return AVAILABLE_SOLC_VERSIONS, installed_versions


def _get_unique_build_json(
    output_evm: dict,
    contract_node: Any,
    stmt_nodes: StatementNodes,
    branch_nodes: BranchNodes,
    has_fallback: bool,
) -> SolidityBuildJson:
    paths = {
        str(i.contract_id): i.parent().absolutePath
        for i in [contract_node] + contract_node.dependencies
    }

    bytecode = _format_link_references(output_evm)
    bytecode_json: DeployedBytecodeJson = output_evm["deployedBytecode"]

    without_metadata = _remove_metadata(bytecode_json["object"])
    instruction_count = len(without_metadata) // 2

    pc_map, statement_map, branch_map = _generate_coverage_data(
        bytecode_json["sourceMap"],
        bytecode_json["opcodes"],
        contract_node,
        stmt_nodes,
        branch_nodes,
        has_fallback,
        instruction_count,
    )

    dependencies = []
    for node in contract_node.dependencies:
        if node.nodeType == "ContractDefinition":
            # use contract aliases when recording dependencies, to avoid
            # potential namespace collisions when importing across projects
            name = node.name
            path_str = node.parent().absolutePath
            dependencies.append(_get_alias(name, path_str))

    # This is not a complete SolidityBuildJson object but that's okay for now
    # TODO: make a partial SolidityBuildJson subclass for this dtype
    build_json: SolidityBuildJson = {  # type: ignore [typeddict-item]
        "allSourcePaths": paths,
        "bytecode": bytecode,
        "bytecodeSha1": sha1(_remove_metadata(bytecode).encode()).hexdigest(),  # type: ignore [typeddict-item]
        "coverageMap": {"statements": statement_map, "branches": branch_map},
        "dependencies": dependencies,
        "offset": contract_node.offset,
        "pcMap": pc_map,
        "type": contract_node.contractKind,
    }
    return build_json


def _format_link_references(evm: dict) -> HexStr:
    # Standardizes formatting for unlinked libraries within bytecode
    bytecode_json: BytecodeJson = evm["bytecode"]
    bytecode = bytecode_json["object"]
    link_refs: dict[str, dict] = bytecode_json.get("linkReferences", {})
    references = ((k, x) for v in link_refs.values() for k, x in v.items())
    for n, loc in ((i[0], x["start"] * 2) for i in references for x in i[1]):
        bytecode = f"{bytecode[:loc]}__{n[:36]:_<36}__{bytecode[loc+40:]}"  # type: ignore [assignment]
    return bytecode


def _remove_metadata(bytecode: HexStr) -> HexStr:
    return bytecode[: -(int(bytecode[-4:], 16) + 2) * 2] if bytecode else ""  # type: ignore [return-value]


def _generate_coverage_data(
    source_map_str: str,
    opcodes_str: str,
    contract_node: Any,
    stmt_nodes: StatementNodes,
    branch_nodes: BranchNodes,
    has_fallback: bool,
    instruction_count: int,
) -> tuple[PCMap, StatementMap, BranchMap]:
    # Generates data used by Brownie for debugging and coverage evaluation
    if not opcodes_str:
        return PCMap({}), {}, {}

    source_map = deque(expand_source_map(source_map_str))
    opcodes = deque(opcodes_str.split(" "))

    contract_nodes = [contract_node] + contract_node.dependencies
    source_nodes: dict[IntegerString, NodeBase] = {
        str(i.contract_id): i.parent() for i in contract_nodes
    }

    stmt_nodes = {i: stmt_nodes[i].copy() for i in source_nodes}
    statement_map: StatementMap = {i: {} for i in source_nodes}

    # possible branch offsets
    branch_original = {i: branch_nodes[i].copy() for i in source_nodes}
    branch_nodes = {i: {i.offset for i in branch_nodes[i]} for i in source_nodes}
    # currently active branches, awaiting a jumpi
    branch_active: dict[str, dict[Offset, int]] = {i: {} for i in source_nodes}
    # branches that have been set
    branch_set: dict[IntegerString, dict[Offset, tuple[int, int]]] = {i: {} for i in source_nodes}

    count, pc = 0, 0
    pc_list: PcList = []
    revert_map: dict[tuple[IntegerString, Offset], list[int]] = {}
    fallback_hexstr: str = "unassigned"

    optimizer_revert = get_version() < Version("0.8.0")

    active_source_node: NodeBase | None = None
    active_fn_node: NodeBase | None = None
    active_fn_name: str | None = None
    first_source = source_map[0]

    while source_map and source_map[-1][2] == -1:
        # trim the end of the source map where there are no contracts associated
        # this is required because sometimes the source map is too long
        # likely a side effect of the YUL optimizer ¯\_(ツ)_/¯
        source_map.pop()

    while source_map:
        # format of source_map is [start, stop, contract_id, jump code]
        source = source_map.popleft()
        start, stop, contract_id_int, jump_code = source

        op = opcodes.popleft()
        this: ProgramCounter = {"op": op, "pc": pc}  # type: ignore [typeddict-item]
        pc_list.append(this)

        if (
            has_fallback is False
            and fallback_hexstr == "unassigned"
            and op == "REVERT"
            and [i["op"] for i in pc_list[-4:-1]] == ["JUMPDEST", "PUSH1", "DUP1"]
        ):
            # flag the REVERT op at the end of the function selector,
            # later reverts may jump to it instead of having their own REVERT op
            fallback_hexstr = f"0x{hex(pc - 4).upper()[2:]}"
            this["first_revert"] = True

        if jump_code != "-":
            this["jump"] = jump_code

        pc += 1
        if op.startswith("PUSH") and opcodes[0][:2] == "0x":
            this["value"] = opcodes.popleft()
            pc += int(op[4:])

        # for REVERT opcodes without a source offset, try to infer one
        if (contract_id_int == -1 or source == first_source) and op == "REVERT":
            _find_revert_offset(
                pc_list, source_map, active_source_node, active_fn_node, active_fn_name
            )
        if contract_id_int == -1:
            continue

        # set contract path (-1 means none)
        contract_id = str(source[2])
        if contract_id not in source_nodes:
            # In Solidity >=0.7.2 contract ID can reference an AST within the YUL-optimization
            # "generatedSources". Brownie does not support coverage evaluation within these
            # sources, so we consider to this to be unmapped.
            continue

        active_source_node = source_nodes[contract_id]
        this["path"] = contract_id

        # set source offset (-1 means none)
        if start == -1:
            continue
        offset: Offset = (start, start + stop)  # type: ignore [assignment]
        this["offset"] = offset

        if op == "REVERT" and not optimizer_revert:
            # In Solidity >=0.8.0, an optimization is applied to reverts with an error string
            # such that all reverts appear to happen at the same point in the source code.
            # We mark this REVERT as the "optimizer revert" so that when it's encountered in
            # a trace we know to look back to find the actual revert location.
            if fn_node := active_source_node.children(
                include_parents=False,
                include_children=True,
                required_offset=offset,
                filters=(
                    {"nodeType": "FunctionCall", "expression.name": "revert"},
                    {"nodeType": "FunctionCall", "expression.name": "require"},
                ),
            ):
                args = len(fn_node[0].arguments)
                if args == 2 or (fn_node[0].expression.name == "revert" and args):
                    optimizer_revert = True
                    this["optimizer_revert"] = True

        # add error messages for INVALID opcodes
        if op == "INVALID":
            _set_invalid_error_string(active_source_node, this)

        # for JUMPI instructions, set active branch markers
        if branch_active[contract_id] and op == "JUMPI":
            for offset in branch_active[contract_id]:
                # ( program counter index, JUMPI index)
                branch_set[contract_id][offset] = (
                    branch_active[contract_id][offset],
                    len(pc_list) - 1,
                )
            branch_active[contract_id].clear()

        # if op relates to previously set branch marker, clear it
        elif offset in branch_nodes[contract_id]:
            if offset in branch_set[contract_id]:
                del branch_set[contract_id][offset]
            branch_active[contract_id][offset] = len(pc_list) - 1

        try:
            # set fn name and statement coverage marker
            if "offset" in pc_list[-2] and offset == pc_list[-2]["offset"]:
                this["fn"] = active_fn_name  # type: ignore [typeddict-item]
            else:
                active_fn_node, active_fn_name = _get_active_fn(active_source_node, offset)  # type: ignore [arg-type]
                this["fn"] = active_fn_name
                stmt_offset: Offset = next(
                    i for i in stmt_nodes[contract_id] if sources.is_inside_offset(offset, i)
                )
                stmt_nodes[contract_id].discard(stmt_offset)
                statement_map[contract_id].setdefault(active_fn_name, {})[count] = stmt_offset
                this["statement"] = count
                count += 1
        except (KeyError, IndexError, StopIteration):
            pass

        if this.get("value", None) == fallback_hexstr and opcodes[0] in ("JUMP", "JUMPI"):
            # track all jumps to the initial revert
            key = (this["path"], this["offset"])
            revert_map.setdefault(key, []).append(len(pc_list))

    while opcodes[0] not in ("INVALID", "STOP") and pc < instruction_count:
        # necessary because sometimes solidity returns an incomplete source map
        op = opcodes.popleft()
        this = {"op": op, "pc": pc}  # type: ignore [typeddict-item]
        pc_list.append(this)
        pc += 1
        if op.startswith("PUSH") and opcodes[0][:2] == "0x":
            this["value"] = opcodes.popleft()
            pc += int(op[4:])

    # compare revert and require statements against the map of revert jumps
    for (contract_id, fn_offset), values in revert_map.items():
        fn_node = source_nodes[contract_id].children(
            depth=2,
            include_children=False,
            required_offset=fn_offset,
            filters={"nodeType": "FunctionDefinition"},
        )
        if len(fn_node) == 0:
            # In Solidity >=0.8.13, with the viaIR option set, there is a dispatch
            # function present in the generated bytecode
            continue
        revert_nodes = fn_node[0].children(
            filters=(
                {"nodeType": "FunctionCall", "expression.name": "revert"},
                {"nodeType": "FunctionCall", "expression.name": "require"},
            )
        )
        for node in revert_nodes:
            offset = node.offset
            # if the node offset is not in the source map, apply it's offset to the JUMPI op
            if not any("offset" in x and x["offset"] == offset for x in pc_list):
                pc_list[values[0]].update(offset=offset, jump_revert=True)  # type: ignore [call-arg]
                del values[0]

    # set branch index markers and build final branch map
    branch_map: BranchMap = {i: {} for i in source_nodes}
    for path, markers in branch_set.items():
        for offset, idx in markers.items():
            # for branch to be hit, need an op relating to the source and the next JUMPI
            # this is because of how the compiler optimizes nested BinaryOperations
            if "fn" in pc_list[idx[0]]:
                fn = pc_list[idx[0]]["fn"]
                pc_list[idx[0]]["branch"] = count
                pc_list[idx[1]]["branch"] = count
                node = next(i for i in branch_original[path] if i.offset == offset)
                branch_map[path].setdefault(fn, {})[count] = offset + (node.jump,)  # type: ignore [arg-type]
                count += 1

    pc_map = PCMap({i.pop("pc"): i for i in pc_list})
    return pc_map, statement_map, branch_map


def _find_revert_offset(
    pc_list: PcList,
    source_map: Deque[Source],
    source_node: NodeBase,
    fn_node: NodeBase,
    fn_name: str | None,
) -> None:
    # attempt to infer a source offset for reverts that do not have one

    if source_map:
        # is not the last instruction
        if len(pc_list) >= 8 and pc_list[-8]["op"] == "CALLVALUE":
            # reference to CALLVALUE 8 instructions previous is a nonpayable function check
            pc_list[-1].update(  # type: ignore [call-arg]
                dev="Cannot send ether to nonpayable function",
                fn=pc_list[-8].get("fn", "<unknown>"),
                offset=pc_list[-8].get("offset"),
                path=pc_list[-8].get("path"),
            )
            return

    # if there is active function, we are still in the function selector table
    if not fn_node:
        return

    # get the offset of the next instruction
    next_offset = None
    if source_map:
        next_instruction = source_map[0]
        if next_instruction[2] != -1:
            next_start = next_instruction[0]
            next_stop = next_start + next_instruction[1]
            next_offset = (next_start, next_stop)

    # if the next instruction offset is not equal to the offset of the active function,
    # but IS contained within the active function, apply this offset to the current
    # instruction

    fn_node_offset: Offset = fn_node.offset
    if (
        next_offset
        and next_offset != fn_node_offset
        and is_inside_offset(next_offset, fn_node_offset)
    ):
        pc_list[-1].update(path=str(source_node.contract_id), fn=fn_name, offset=next_offset)  # type: ignore [call-arg]
        return

    # if any of the previous conditions are not satisfied, this is the final revert
    # statement within a function
    if fn_node[-1].nodeType == "ExpressionStatement":
        expr = fn_node[-1].expression

        if expr.nodeType == "FunctionCall" and expr.get("expression.name") in ("revert", "require"):
            pc_list[-1].update(  # type: ignore [call-arg]
                path=str(source_node.contract_id), fn=fn_name, offset=expr.expression.offset
            )


def _set_invalid_error_string(source_node: NodeBase, pc_map: ProgramCounter) -> None:
    # set custom error string for INVALID opcodes
    try:
        node = source_node.children(include_children=False, offset_limits=pc_map["offset"])[0]
    except IndexError:
        return
    node_type: str = node.nodeType
    if node_type == "IndexAccess":
        pc_map["dev"] = "Index out of range"
    elif node_type == "BinaryOperation":
        operator: str = node.operator
        if operator == "/":
            pc_map["dev"] = "Division by zero"
        elif operator == "%":
            pc_map["dev"] = "Modulus by zero"


def _get_active_fn(source_node: NodeBase, offset: tuple[int, int]) -> tuple[NodeBase, str]:
    fn_node = source_node.children(
        depth=2, required_offset=offset, filters={"nodeType": "FunctionDefinition"}
    )[0]
    name = getattr(fn_node, "name", None)
    if not name:
        if getattr(fn_node, "kind", "function") != "function":
            name = f"<{fn_node.kind}>"
        elif getattr(fn_node, "isConstructor", False):
            name = "<constructor>"
        else:
            name = "<fallback>"

    parent = fn_node.parent()
    if parent.nodeType == "SourceUnit":
        # the function exists outside a contract
        return fn_node, name

    return fn_node, f"{fn_node.parent().name}.{name}"


def _get_nodes(output_json: dict) -> tuple[list[NodeBase], StatementNodes, BranchNodes]:
    source_nodes: list[NodeBase] = solcast.from_standard_output(output_json)
    stmt_nodes = _get_statement_nodes(source_nodes)
    branch_nodes = _get_branch_nodes(source_nodes)
    return source_nodes, stmt_nodes, branch_nodes


def _get_statement_nodes(source_nodes: list[NodeBase]) -> StatementNodes:
    # Given a list of source nodes, returns a dict of lists of statement nodes
    return {
        str(node.contract_id): {
            i.offset
            for i in node.children(
                include_parents=False,
                filters={"baseNodeType": "Statement"},
                exclude_filter={"isConstructor": True},
            )
        }
        for node in source_nodes
    }


def _get_branch_nodes(source_nodes: list[NodeBase]) -> BranchNodes:
    # Given a list of source nodes, returns a dict of lists of nodes corresponding
    # to possible branches in the code
    branches: BranchNodes = {}
    for node in source_nodes:
        contract_branches: set[NodeBase] = set()
        branches[str(node.contract_id)] = contract_branches
        for contract_node in node.children(depth=1, filters={"nodeType": "ContractDefinition"}):
            for i in contract_node:
                for child_node in i.children(
                    filters=(
                        {"nodeType": "FunctionCall", "expression.name": "require"},
                        {"nodeType": "IfStatement"},
                        {"nodeType": "Conditional"},
                    )
                ):
                    contract_branches |= _get_recursive_branches(child_node)
    return branches


def _get_recursive_branches(base_node: NodeBase) -> set[NodeBase]:
    node_type = base_node.nodeType

    # if node is IfStatement or Conditional, look only at the condition
    node = base_node if node_type == "FunctionCall" else base_node.condition
    # for IfStatement, jumping indicates evaluating false
    jump_is_truthful = node_type != "IfStatement"

    filters = (
        {"nodeType": "BinaryOperation", "typeDescriptions.typeString": "bool", "operator": "||"},
        {"nodeType": "BinaryOperation", "typeDescriptions.typeString": "bool", "operator": "&&"},
    )
    all_binaries = node.children(include_parents=True, include_self=True, filters=filters)

    # if no BinaryOperation nodes are found, this node is the branch
    if not all_binaries:
        # if node is FunctionCall, look at the first argument
        if node_type == "FunctionCall":
            node = node.arguments[0]
        # some versions of solc do not map IfStatement unary operations to bytecode
        elif node.nodeType == "UnaryOperation":
            node = node.subExpression
        node.jump = jump_is_truthful
        return {node}

    # look at children of BinaryOperation nodes to find all possible branches
    binary_branches = set()
    for node in (x for i in all_binaries for x in (i.leftExpression, i.rightExpression)):
        if node.children(include_self=True, filters=filters):
            continue
        _jump = jump_is_truthful
        if not _is_rightmost_operation(node, base_node.depth):
            _jump = _check_left_operator(node, base_node.depth)
        if node.nodeType == "UnaryOperation":
            node = node.subExpression
        node.jump = _jump
        binary_branches.add(node)

    return binary_branches


def _is_rightmost_operation(node: NodeBase, depth: int) -> bool:
    # Check if the node is the final operation within the expression
    parents = node.parents(depth, _BINOPS_PARAMS)
    return not any(i.leftExpression == node or node.is_child_of(i.leftExpression) for i in parents)


def _check_left_operator(node: NodeBase, depth: int) -> bool:
    # Find the nearest parent boolean where this node sits on the left side of
    # the comparison, and return True if that node's operator is ||
    parents = node.parents(depth, _BINOPS_PARAMS)

    op = next(
        i for i in parents if i.leftExpression == node or node.is_child_of(i.leftExpression)
    ).operator
    return op == "||"



"""
Brownie is written in Python, but compiled to C using mypyc.

This file exists to hold C constants for commonly used external imports.

C constants let us bypass reference counting and are significantly more efficient to access.

Check out https://github.com/mypyc/mypyc for more info.
"""

import collections
import copy
import decimal
import hashlib
import importlib
import pathlib
import re
from typing import Final

import faster_eth_utils.toolz
import faster_hexbytes
import semantic_version
import ujson

# BUILTINS
# collections
defaultdict: Final = collections.defaultdict
deque: Final = collections.deque

# copy
deepcopy: Final = copy.deepcopy

# decimal
Decimal: Final = decimal.Decimal
getcontext: Final = decimal.getcontext

# hashlib
sha1: Final = hashlib.sha1

# importlib
import_module: Final = importlib.import_module

# pathlib
Path: Final = pathlib.Path

# re
regex_compile: Final = re.compile
regex_findall: Final = re.findall
regex_finditer: Final = re.finditer
regex_fullmatch: Final = re.fullmatch
regex_match: Final = re.match
regex_sub: Final = re.sub


# DEPENDENCIES
# faster_hexbytes
HexBytes: Final = faster_hexbytes.HexBytes

# semantic_version
NpmSpec: Final = semantic_version.NpmSpec
Version: Final = semantic_version.Version

# toolz
mapcat: Final = faster_eth_utils.toolz.mapcat

# ujson
ujson_dump: Final = ujson.dump
ujson_dumps: Final = ujson.dumps
ujson_load: Final = ujson.load
ujson_loads: Final = ujson.loads



import functools
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from typing import Any, Dict, Final, List, Optional, Type

from web3 import Web3
from web3.types import RPCEndpoint

partial: Final = functools.partial


class BrownieMiddlewareABC(ABC):
    """
    Base ABC for all middlewares.

    This class must be inherited in order for a middleware to be discovered
    and added to the `web3` object upon connecting to a network.
    """

    def __init__(self, w3: Web3) -> None:
        """
        Initialize the middleware.

        Subclasses may optionally include this method. It is called only once,
        when the middleware is being added.
        """
        self.w3: Final = w3

    @classmethod
    @abstractmethod
    def get_layer(cls, w3: Web3, network_type: str) -> int | None:
        """
        Return the target layer of this middleware.

        All builtin middlewares are considered to be on layer 0. Middlewares are called in
        ascending order prior to the request, and descending order after the request.
        """
        raise NotImplementedError

    def __call__(self, make_request: Callable, w3: Web3) -> Callable:
        """
        Receive the initial middleware request and return `process_request`.

        Subclasses should NOT include this method.
        """
        return partial(self.process_request, make_request)

    @abstractmethod
    def process_request(
        self,
        make_request: Callable,
        method: RPCEndpoint,
        params: Sequence[Any],
    ) -> dict[str, Any]:
        """
        Process an RPC request.

        Data is pre-processed, sent onward via `make_request`, processed further and returned.
        See https://web3py.readthedocs.io/en/stable/internals.html#middlewares for more info.

        Arguments
        ---------
        method : str
            A string representing the JSON-RPC method that is being called.
        params : List
            An iterable of the parameters for the JSON-RPC method being called.

        Returns
        -------
        Dict
            A dictionary containing either a 'result' key in the case
            of success, or an 'error' key in the case of failure.
        """
        raise NotImplementedError

    def uninstall(self) -> None:
        """
        Uninstall a middleware.

        Subclasses may optionally include this method if any cleanup is required
        when they are removed. Note that you must not assume network connectivity
        when this method is called.
        """


def get_middlewares(web3: Web3, network_type: str) -> dict:
    """
    Get a list of middlewares to be used for the given web3 object.

    Arguments
    ---------
    web3 : Web3
        The active web3 object, already connected to the current network.
    network_type : str
        One of "live" or "development".
    """
    middleware_layers: dict[int, list[type[BrownieMiddlewareABC]]] = {}
    for obj in _middlewares:
        layer = obj.get_layer(web3, network_type)
        if layer is not None:
            middleware_layers.setdefault(layer, []).append(obj)

    return middleware_layers


# this must go down here to prevent a circ import issue
from brownie.network.middlewares._setup import load_middlewares  # noqa: E402

_middlewares: Final = load_middlewares()



#!/usr/bin/python3

import os
import time
from pathlib import Path

from ens import ENS
from eth_typing import ChecksumAddress, HexStr
from requests import HTTPError
from ujson import JSONDecodeError
from web3 import HTTPProvider, IPCProvider
from web3 import Web3 as _Web3
from web3 import WebsocketProvider
from web3.contract.contract import ContractEvent  # noqa
from web3.contract.contract import ContractEvents as _ContractEvents  # noqa
from web3.gas_strategies.rpc import rpc_gas_price_strategy

from brownie._c_constants import ujson_dump, ujson_load
from brownie._config import CONFIG, _get_data_folder
from brownie.convert import to_address
from brownie.exceptions import MainnetUndefined, UnsetENSName
from brownie.network.middlewares import get_middlewares

_chain_uri_cache: dict = {}


class Web3(_Web3):
    """Brownie Web3 subclass"""

    def __init__(self) -> None:
        super().__init__(HTTPProvider("null"))
        self.provider = None
        self._mainnet_w3: _Web3 | None = None
        self._genesis_hash: HexStr | None = None
        self._chain_uri: str | None = None
        self._custom_middleware: set = set()
        self._supports_traces = None
        self._chain_id: int | None = None

    def _remove_middlewares(self) -> None:
        for middleware in self._custom_middleware:
            try:
                self.middleware_onion.remove(middleware)
            except ValueError:
                pass
            middleware.uninstall()
        self._custom_middleware.clear()

    def connect(self, uri: str, timeout: int = 30) -> None:
        """Connects to a provider"""
        self._remove_middlewares()
        self.provider = None

        uri = _expand_environment_vars(uri)
        try:
            if Path(uri).exists():
                self.provider = IPCProvider(uri, timeout=timeout)
        except OSError:
            pass

        if self.provider is None:
            if uri.startswith("ws"):
                self.provider = WebsocketProvider(uri, {"close_timeout": timeout})
            elif uri.startswith("http"):

                self.provider = HTTPProvider(uri, {"timeout": timeout})
            else:
                raise ValueError(
                    "Unknown URI - must be a path to an IPC socket, a websocket "
                    "beginning with 'ws' or a URL beginning with 'http'"
                )

        try:
            if self.isConnected():
                self.reset_middlewares()
        except Exception:
            # checking an invalid connection sometimes raises on windows systems
            pass

    def reset_middlewares(self) -> None:
        """
        Uninstall and reinject all custom middlewares.
        """
        if self.provider is None:
            raise ConnectionError("web3 is not currently connected")
        self._remove_middlewares()

        middleware_layers = get_middlewares(self, CONFIG.network_type)

        # middlewares with a layer below zero are injected
        to_inject = sorted((i for i in middleware_layers if i < 0), reverse=True)
        for layer in to_inject:
            for obj in middleware_layers[layer]:
                middleware = obj(self)
                self.middleware_onion.inject(middleware, layer=0)
                self._custom_middleware.add(middleware)

        # middlewares with a layer of zero or greater are added
        to_add = sorted(i for i in middleware_layers if i >= 0)
        for layer in to_add:
            for obj in middleware_layers[layer]:
                middleware = obj(self)
                self.middleware_onion.add(middleware)
                self._custom_middleware.add(middleware)

    def disconnect(self) -> None:
        """Disconnects from a provider"""
        if self.provider:
            self.provider = None
            self._genesis_hash = None
            self._chain_uri = None
            self._supports_traces = None
            self._chain_id = None
            self._remove_middlewares()

    def is_connected(self) -> bool:
        return super().is_connected() if self.provider else False

    def isConnected(self) -> bool:
        # retained to avoid breaking an interface explicitly defined in brownie
        return self.is_connected()

    @property
    def supports_traces(self) -> bool:
        if not self.provider:
            return False

        # Send a malformed request to `debug_traceTransaction`. If the error code
        # returned is -32601 "endpoint does not exist/is not available" we know
        # traces are not possible. Any other error code means the endpoint is open.
        if self._supports_traces is None:
            try:
                response = self.provider.make_request("debug_traceTransaction", [])
                self._supports_traces = response["error"]["code"] != -32601
            except HTTPError:
                self._supports_traces = False

        return self._supports_traces

    @property
    def _mainnet(self) -> _Web3:
        # a web3 instance connected to the mainnet
        if self.is_connected() and CONFIG.active_network["id"] == "mainnet":
            return self
        try:
            mainnet = CONFIG.networks["mainnet"]
        except KeyError:
            raise MainnetUndefined("No 'mainnet' network defined") from None
        if not self._mainnet_w3:
            uri = _expand_environment_vars(mainnet["host"])
            self._mainnet_w3 = _Web3(HTTPProvider(uri))
        return self._mainnet_w3

    @property
    def genesis_hash(self) -> HexStr:
        """The genesis hash of the currently active network."""
        if self.provider is None:
            raise ConnectionError("web3 is not currently connected")
        if self._genesis_hash is None:
            # removeprefix is used for compatibility with both hexbytes<1 and >=1
            self._genesis_hash = HexStr(self.eth.get_block(0)["hash"].hex().removeprefix("0x"))
        return self._genesis_hash

    @property
    def chain_uri(self) -> str:
        if self.provider is None:
            raise ConnectionError("web3 is not currently connected")
        if self.genesis_hash not in _chain_uri_cache:
            block_number = max(self.eth.block_number - 16, 0)
            # removeprefix is used for compatibility with both hexbytes<1 and >=1
            block_hash = self.eth.get_block(block_number)["hash"].hex().removeprefix("0x")
            chain_uri = f"blockchain://{self.genesis_hash}/block/{block_hash}"
            _chain_uri_cache[self.genesis_hash] = chain_uri
        return _chain_uri_cache[self.genesis_hash]

    @property
    def chain_id(self) -> int:
        # chain ID is needed each time we a sign a transaction, however we
        # cache it after the first request to avoid redundant RPC calls
        if self.provider is None:
            raise ConnectionError("web3 is not currently connected")
        if self._chain_id is None:
            self._chain_id = self.eth.chain_id
        return self._chain_id


def _expand_environment_vars(uri: str) -> str:
    if "$" not in uri:
        return uri
    expanded = os.path.expandvars(uri)
    if uri != expanded:
        return expanded
    raise ValueError(f"Unable to expand environment variable in host setting: '{uri}'")


def _get_path() -> Path:
    return _get_data_folder().joinpath("ens.json")


def _resolve_address(domain: str) -> ChecksumAddress:
    # convert ENS domain to address
    if not isinstance(domain, str) or "." not in domain:
        return to_address(domain)
    domain = domain.lower()
    if domain not in _ens_cache or time.time() - _ens_cache[domain][1] > 86400:
        try:
            ns = ENS.from_web3(web3._mainnet)
        except MainnetUndefined as e:
            raise MainnetUndefined(f"Cannot resolve ENS address - {e}") from None
        address = ns.address(domain)
        _ens_cache[domain] = [address, int(time.time())]
        with _get_path().open("w") as fp:
            ujson_dump(_ens_cache, fp)
    if _ens_cache[domain][0] is None:
        raise UnsetENSName(f"ENS domain '{domain}' is not set")
    return _ens_cache[domain][0]


web3 = Web3()
web3.eth.set_gas_price_strategy(rpc_gas_price_strategy)

try:
    with _get_path().open() as fp:
        _ens_cache: dict = ujson_load(fp)
except (FileNotFoundError, JSONDecodeError):
    _ens_cache = {}



import re
from collections.abc import Iterator
from typing import Any, DefaultDict, Final, final

from brownie._c_constants import Path, defaultdict, mapcat, regex_compile, regex_sub
from brownie.utils.toposort import toposort_flatten

# Pattern matching Solidity `import-directive`, capturing path component
# https://docs.soliditylang.org/en/latest/grammar.html#a4.SolidityParser.importDirective
IMPORT_PATTERN: Final = regex_compile(
    r"(?<=\n)?import(?P<prefix>.*)(?P<quote>[\"'])(?P<path>.*)(?P=quote)(?P<suffix>.*)(?=\n)"
)
PRAGMA_PATTERN: Final = regex_compile(r"^pragma.*;$", re.MULTILINE)
LICENSE_PATTERN: Final = regex_compile(r"^// SPDX-License-Identifier: (.*)$", re.MULTILINE)


@final
class Flattener:
    """Brownie's Robust Solidity Flattener."""

    def __init__(
        self,
        primary_source_fp: str,
        contract_name: str,
        remappings: dict[str, str],
        compiler_settings: dict[str, Any],
    ) -> None:
        self.sources: Final[dict[str, str]] = {}
        self.dependencies: Final[DefaultDict[str, set[str]]] = defaultdict(set)
        self.compiler_settings: Final = compiler_settings
        self.contract_name: Final = contract_name
        self.contract_file: Final = self.path_to_name(primary_source_fp)
        self.remappings: Final = remappings

        self.traverse(primary_source_fp)

        license_search = _LICENSE_PATTERN_SEARCH(self.path_to_name(primary_source_fp))
        self.license: Final = license_search.group(1) if license_search else "NONE"

    @classmethod
    def path_to_name(cls, pth: str) -> str:
        """Turn the full-path of every Solidity file to a unique shorten name.

        Note, that sometimes there could be several different files with the same name in a project,
        so these files should keep uniq name to correct verification.
        """
        return "contracts/" + pth.split("/contracts/")[1]

    def traverse(self, fp: str) -> None:
        """Traverse a contract source files dependencies.

        Files are read in, import statement path components are substituted for their absolute
        path, and the modified source is saved along with it's dependencies.

        Args:
            fp: The contract source file to traverse, if it's already been traversed, return early.
        """
        # if already traversed file, return early
        name = self.path_to_name(fp)
        fp_obj = Path(fp)
        if name in self.sources:
            return

        # read in the source file
        source = fp_obj.read_text()

        # path sanitization lambda fn
        sanitize = lambda path: self.make_import_absolute(  # noqa: E731
            self.remap_import(path), fp_obj.parent.as_posix()
        )
        # replacement function for re.sub, we just sanitize the path
        repl = (  # noqa: E731
            lambda m: f'import{m.group("prefix")}'
            + f'"{self.path_to_name(sanitize(m.group("path")))}"'
            + f'{m.group("suffix")}'
        )
        self.sources[name] = _IMPORT_PATTERN_SUB(repl, source)
        if fp_obj.name not in self.dependencies:
            self.dependencies[name] = set()

        # traverse dependency files - can circular imports happen?
        for m in _IMPORT_PATTERN_FINDITER(source):
            import_path = sanitize(m.group("path"))
            self.dependencies[name].add(self.path_to_name(import_path))
            self.traverse(import_path)

    @property
    def flattened_source(self) -> str:
        """The flattened source code for use verifying."""
        flattened_deps = toposort_flatten(self.dependencies)
        # all source files in the correct order for concatenation
        sources = [self.sources[x] for x in flattened_deps]

        pragmas_iter: Iterator[str] = mapcat(_PRAGMA_PATTERN_FINDALL, sources)
        # all pragma statements, we already have the license used + know which compiler
        # version is used via the build info
        pragmas = {s.strip() for s in pragmas_iter}

        # now we go through and remove all imports/pragmas/license stuff, then flatten
        flat = (
            "\n".join(pragma for pragma in pragmas if "pragma solidity" not in pragma)
            + "\n\n"
            + "\n".join(
                f"// File: {file}\n\n{_wipe(src)}" for src, file in zip(sources, flattened_deps)
            )
        )

        # hopefully this doesn't mess up anything pretty, but just gotta remove all
        # that extraneous whitespace
        return regex_sub(r"\n{3,}", "\n\n", flat)

    @property
    def standard_input_json(self) -> dict:
        """Useful for etherscan verification via solidity-standard-json-input mode.

        Sadly programmatic upload of this isn't available at the moment (2021-10-11)
        """
        return {
            "language": "Solidity",
            "sources": {k: {"content": v} for k, v in self.sources.items()},
            "settings": self.compiler_settings,
        }

    def remap_import(self, import_path: str) -> str:
        """Remap imports in a solidity source file.

        Args:
            import_path: The path component of an import directive from a solidity source file.

        Returns:
            str: The import path string correctly remapped.
        """
        for k, v in self.remappings.items():
            if import_path.startswith(k):
                return import_path.replace(k, v, 1)
        return import_path

    @staticmethod
    def make_import_absolute(import_path: str, source_file_dir: str) -> str:
        """Make an import path absolute, if it is not already.

        Args:
            source_file_dir: The parent directory of the source file where the import path appears.
            import_path: The path component of an import directive (should already remapped).

        Returns:
            str: The import path string in absolute form.
        """
        path = Path(import_path)
        if path.is_absolute():
            return path.as_posix()

        dir_path = Path(source_file_dir).resolve()
        newpath = (dir_path / path).resolve()
        while not newpath.exists():
            dir_path = dir_path.parent
            newpath = (dir_path / path).resolve()
            if dir_path == Path("/"):
                raise FileNotFoundError(f"Cannot determine location of {import_path}")
        return newpath.as_posix()


def _wipe(src: str) -> str:
    """go through and remove all imports/pragmas/license stuff."""
    return _PRAGMA_PATTERN_SUB("", _LICENSE_PATTERN_SUB("", _IMPORT_PATTERN_SUB("", src)))


# Internal C Constants

_IMPORT_PATTERN_FINDITER: Final = IMPORT_PATTERN.finditer
_IMPORT_PATTERN_SUB: Final = IMPORT_PATTERN.sub

_PRAGMA_PATTERN_FINDALL: Final = PRAGMA_PATTERN.findall
_PRAGMA_PATTERN_SUB: Final = PRAGMA_PATTERN.sub

_LICENSE_PATTERN_SEARCH: Final = LICENSE_PATTERN.search
_LICENSE_PATTERN_SUB: Final = LICENSE_PATTERN.sub



#!/usr/bin/python3

import asyncio
import io
import os
import time
import warnings
from collections.abc import Callable, Coroutine, Iterator
from pathlib import Path
from re import Match
from textwrap import TextWrapper
from threading import get_ident  # noqa
from typing import TYPE_CHECKING, Any, Final, Optional, Union

import requests
import solcx
from eth_typing import ABIConstructor, ABIElement, ABIFunction, ChecksumAddress, HexAddress, HexStr
from faster_eth_abi import decode as decode_abi
from faster_eth_abi import encode as encode_abi
from faster_eth_utils import combomethod
from vvm import get_installable_vyper_versions
from vvm.utils.convert import to_vyper_version
from web3._utils import filters
from web3.datastructures import AttributeDict
from web3.types import LogReceipt

from brownie._c_constants import (
    HexBytes,
    Version,
    regex_findall,
    ujson_dump,
    ujson_dumps,
    ujson_load,
    ujson_loads,
)
from brownie._config import BROWNIE_FOLDER, CONFIG, REQUEST_HEADERS, _load_project_compiler_config
from brownie.convert.datatypes import Wei
from brownie.convert.normalize import format_input, format_output
from brownie.convert.utils import (
    build_function_selector,
    build_function_signature,
    get_type_strings,
)
from brownie.exceptions import (
    BrownieCompilerWarning,
    BrownieEnvironmentWarning,
    ContractExists,
    ContractNotFound,
    UndeployedLibrary,
    VirtualMachineError,
    decode_typed_error,
    parse_errors_from_abi,
)
from brownie.project import compiler
from brownie.project.flattener import Flattener
from brownie.typing import (
    AccountsType,
    ContractBuildJson,
    ContractName,
    FunctionName,
    Language,
    Selector,
    TransactionReceiptType,
)
from brownie.utils import color, hexbytes_to_hexstring
from brownie.utils._color import bright_blue, bright_green, bright_magenta, bright_red

from . import accounts, chain
from .event import _add_deployment_topics, _get_topics, event_watcher
from .state import (
    _add_contract,
    _add_deployment,
    _find_contract,
    _get_deployment,
    _remove_contract,
    _remove_deployment,
    _revert_register,
)
from .web3 import ContractEvent, _ContractEvents, _resolve_address, web3

if TYPE_CHECKING:
    from brownie.project.main import Project, TempProject

AnyContractMethod = Union["ContractCall", "ContractTx", "OverloadedMethod"]

_unverified_addresses: Final[set[ChecksumAddress]] = set()


class _ContractBase:
    _dir_color: Final = "bright magenta"

    def __init__(
        self,
        project: Union["Project", "TempProject"] | None,
        build: ContractBuildJson,
        sources: dict[str, Any],
    ) -> None:
        self._project = project
        self._build: Final = build.copy()
        self._sources: Final = sources

        abi = self.abi
        self.topics: Final = _get_topics(abi)
        self.selectors: Final[dict[Selector, FunctionName]] = {
            build_function_selector(i): FunctionName(i["name"])
            for i in abi
            if i["type"] == "function"
        }
        # this isn't fully accurate because of overloaded methods - will be removed in `v2.0.0`
        self.signatures: Final[dict[FunctionName, Selector]] = {
            v: k for k, v in self.selectors.items()
        }
        parse_errors_from_abi(abi)

    @property
    def abi(self) -> list[ABIElement]:
        return self._build["abi"]

    @property
    def _name(self) -> ContractName:
        return self._build["contractName"]

    def info(self) -> None:
        """
        Display NatSpec documentation for this contract.
        """
        if natspec := self._build.get("natspec"):
            _print_natspec(natspec)

    def get_method(self, calldata: str) -> str | None:
        sig = calldata[:10].lower()
        return self.selectors.get(sig)

    def decode_input(self, calldata: str | bytes) -> tuple[str, Any]:
        """
        Decode input calldata for this contract.

        Arguments
        ---------
        calldata : str | bytes
            Calldata for a call to this contract

        Returns
        -------
        str
            Signature of the function that was called
        Any
            Decoded input arguments
        """
        if not isinstance(calldata, HexBytes):
            calldata = HexBytes(calldata)

        fn_selector = hexbytes_to_hexstring(calldata[:4])

        abi: ABIFunction | None = None
        for i in self.abi:
            if i["type"] == "function" and build_function_selector(i) == fn_selector:
                abi = i
                break

        if abi is None:
            raise ValueError("Four byte selector does not match the ABI for this contract")

        function_sig = build_function_signature(abi)

        types_list = get_type_strings(abi["inputs"])
        result = decode_abi(types_list, calldata[4:])
        input_args = format_input(abi, result)

        return function_sig, input_args


class ContractContainer(_ContractBase):
    """List-like container class that holds all Contract instances of the same
    type, and is used to deploy new instances of that contract.

    Attributes:
        abi: Complete contract ABI.
        bytecode: Bytecode used to deploy the contract.
        signatures: Dictionary of {'function name': "bytes4 signature"}
        topics: Dictionary of {'event name': "bytes32 topic"}"""

    def __init__(self, project: Any, build: ContractBuildJson) -> None:
        self.tx = None
        self.bytecode: Final = build["bytecode"]
        self._contracts: Final[list["ProjectContract"]] = []
        super().__init__(project, build, project._sources)
        self.deploy: Final = ContractConstructor(self, self._name)
        _revert_register(self)

        # messes with tests if it is created on init
        # instead we create when it's requested, but still define it here
        self._flattener: Flattener = None

    def __iter__(self) -> Iterator["ProjectContract"]:
        return iter(self._contracts)

    def __getitem__(self, i: Any) -> "ProjectContract":
        return self._contracts[i]

    def __delitem__(self, key: Any) -> None:
        item = self._contracts[key]
        self.remove(item)

    def __len__(self) -> int:
        return len(self._contracts)

    def __repr__(self) -> str:
        if CONFIG.argv["cli"] == "console":
            return str(self._contracts)
        return super().__repr__()

    def _reset(self) -> None:
        for contract in self._contracts:
            _remove_contract(contract)
            contract._reverted = True
        self._contracts.clear()

    def _revert(self, height: int) -> None:
        reverted = [
            i
            for i in self._contracts
            if (i.tx and i.tx.block_number is not None and i.tx.block_number > height)
            # removeprefix is used for compatibility with both hexbytes<1 and >=1
            or len(web3.eth.get_code(i.address).hex().removeprefix("0x")) <= 2
        ]
        for contract in reverted:
            self.remove(contract)
            contract._reverted = True

    def remove(self, contract: "ProjectContract") -> None:
        """Removes a contract from the container.

        Args:
            contract: Contract instance of address string of the contract."""
        if contract not in self._contracts:
            raise TypeError("Object is not in container.")
        self._contracts.remove(contract)
        contract._delete_deployment()
        _remove_contract(contract)

    def at(
        self,
        address: HexAddress,
        owner: AccountsType | None = None,
        tx: TransactionReceiptType | None = None,
        persist: bool = True,
    ) -> "ProjectContract":
        """Returns a contract address.

        Raises ValueError if no bytecode exists at the address.

        Args:
            address: Address string of the contract.
            owner: Default Account instance to send contract transactions from.
            tx: Transaction ID of the contract creation."""
        address = _resolve_address(address)
        contract = _find_contract(address)
        if isinstance(contract, ProjectContract):
            if contract._name == self._name and contract._project == self._project:
                return contract
            raise ContractExists(
                f"'{contract._name}' declared at {address} in project '{contract._project._name}'"
            )

        build = self._build
        contract = ProjectContract(self._project, build, address, owner, tx)
        if not _verify_deployed_code(address, build["deployedBytecode"], build["language"]):
            # prevent trace attempts when the bytecode doesn't match
            del contract._build["pcMap"]

        contract._save_deployment()
        _add_contract(contract)
        self._contracts.append(contract)
        if CONFIG.network_type == "live" and persist:
            _add_deployment(contract)

        return contract

    def _add_from_tx(self, tx: TransactionReceiptType) -> None:
        tx._confirmed.wait()
        if tx.status and tx.contract_address is not None:
            try:
                self.at(tx.contract_address, tx.sender, tx)
            except ContractNotFound:
                # if the contract self-destructed during deployment
                pass

    def get_verification_info(self) -> dict:
        """
        Return a dict with flattened source code for this contract
        and further information needed for verification
        """
        language = self._build["language"]
        if language == "Vyper":
            raise TypeError(
                "Etherscan does not support API verification of source code "
                "for vyper contracts. You need to verify the source manually"
            )
        elif language == "Solidity":
            if self._flattener is None:
                source_fp = (
                    Path(self._project._path)
                    .joinpath(self._build["sourcePath"])
                    .resolve()
                    .as_posix()
                )
                config = self._project._compiler_config
                remaps = dict(
                    map(
                        lambda s: s.split("=", 1),
                        compiler._get_solc_remappings(config["solc"]["remappings"]),
                    )
                )
                libs = {lib.strip("_") for lib in regex_findall("_{1,}[^_]*_{1,}", self.bytecode)}
                compiler_settings = {
                    "evmVersion": self._build["compiler"]["evm_version"],
                    "optimizer": config["solc"]["optimizer"],
                    "libraries": {
                        Path(source_fp).name: {lib: self._project[lib][-1].address for lib in libs}
                    },
                }
                self._flattener = Flattener(source_fp, self._name, remaps, compiler_settings)

            build_json = self._build

            return {
                "standard_json_input": self._flattener.standard_input_json,
                "contract_name": build_json["contractName"],
                "compiler_version": build_json["compiler"]["version"],
                "optimizer_enabled": build_json["compiler"]["optimizer"]["enabled"],
                "optimizer_runs": build_json["compiler"]["optimizer"]["runs"],
                "license_identifier": self._flattener.license,
                "bytecode_len": len(build_json["bytecode"]),
            }
        else:
            raise TypeError(f"Unsupported language for source verification: {language}")

    def publish_source(self, contract: "Contract", silent: bool = False) -> bool:
        """Flatten contract and publish source on the selected explorer"""

        api_key = os.getenv("ETHERSCAN_TOKEN")
        if api_key is None:
            raise ValueError(
                "An API token is required to verify contract source code. "
                "Visit https://etherscan.io/register to obtain a token, and "
                "then store it as the environment variable $ETHERSCAN_TOKEN"
            )

        address = _resolve_address(contract.address)

        # Get source code and contract/compiler information
        contract_info = self.get_verification_info()

        # Select matching license code (https://etherscan.io/contract-license-types)
        license_code = 1
        identifier = contract_info["license_identifier"].lower()
        if "unlicensed" in identifier:
            license_code = 2
        elif "mit" in identifier:
            license_code = 3
        elif "agpl" in identifier and "3.0" in identifier:
            license_code = 13
        elif "lgpl" in identifier:
            if "2.1" in identifier:
                license_code = 6
            elif "3.0" in identifier:
                license_code = 7
        elif "gpl" in identifier:
            if "2.0" in identifier:
                license_code = 4
            elif "3.0" in identifier:
                license_code = 5
        elif "bsd-2-clause" in identifier:
            license_code = 8
        elif "bsd-3-clause" in identifier:
            license_code = 9
        elif "mpl" in identifier and "2.0" in identifier:
            license_code = 10
        elif identifier.startswith("osl") and "3.0" in identifier:
            license_code = 11
        elif "apache" in identifier and "2.0" in identifier:
            license_code = 12

        # get constructor arguments
        url = "https://api.etherscan.io/v2/api"
        params_tx: dict = {
            "chainid": web3.chain_id,
            "apikey": api_key,
            "module": "account",
            "action": "txlist",
            "address": address,
            "page": 1,
            "sort": "asc",
            "offset": 1,
        }
        i = 0
        while True:
            response = requests.get(url, params=params_tx, headers=REQUEST_HEADERS)
            if response.status_code != 200:
                raise ConnectionError(
                    f"Status {response.status_code} when querying {url}: {response.text}"
                )
            data = response.json()
            if int(data["status"]) == 1:
                # Constructor arguments received
                break

            # Wait for contract to be recognized by etherscan
            # This takes a few seconds after the contract is deployed
            # After 10 loops we throw with the API result message (includes address)
            if i >= 10:
                raise ValueError(f"API request failed with: {data['result']}")
            elif i == 0 and not silent:
                print(f"Waiting for {url} to process contract...")
            i += 1
            time.sleep(10)

        if data["message"] == "OK":
            constructor_arguments = data["result"][0]["input"][contract_info["bytecode_len"] + 2 :]
        else:
            constructor_arguments = ""

        # Submit verification
        payload_verification: dict = {
            "apikey": api_key,
            "module": "contract",
            "action": "verifysourcecode",
            "contractaddress": address,
            "sourceCode": io.StringIO(ujson_dumps(self._flattener.standard_input_json)),
            "codeformat": "solidity-standard-json-input",
            "contractname": f"{self._flattener.contract_file}:{self._flattener.contract_name}",
            "compilerversion": f"v{contract_info['compiler_version']}",
            "optimizationUsed": 1 if contract_info["optimizer_enabled"] else 0,
            "runs": contract_info["optimizer_runs"],
            # NOTE: This typo is intentional.
            # https://docs.etherscan.io/etherscan-v2/common-verification-errors
            # "There is an easter egg 🐣 on the constructorArguements field spelling,
            # using it as the "correct" spelling may miss your submission!"
            "constructorArguements": constructor_arguments,
            "licenseType": license_code,
        }
        response = requests.post(
            f"{url}?chainid={web3.chain_id}", data=payload_verification, headers=REQUEST_HEADERS
        )
        if response.status_code != 200:
            raise ConnectionError(
                f"Status {response.status_code} when querying {url}: {response.text}"
            )
        data = response.json()
        if int(data["status"]) != 1:
            raise ValueError(f"Failed to submit verification request: {data['result']}")

        # Status of request
        guid = data["result"]
        if not silent:
            print("Verification submitted successfully. Waiting for result...")
        time.sleep(10)
        params_status: dict = {
            "chainid": web3.chain_id,
            "apikey": api_key,
            "module": "contract",
            "action": "checkverifystatus",
            "guid": guid,
        }
        while True:
            response = requests.get(url, params=params_status, headers=REQUEST_HEADERS)
            if response.status_code != 200:
                raise ConnectionError(
                    f"Status {response.status_code} when querying {url}: {response.text}"
                )
            data = response.json()
            if data["result"] == "Pending in queue":
                if not silent:
                    print("Verification pending...")
            else:
                if not silent:
                    color = bright_green if data["message"] == "OK" else bright_red
                    print(f"Verification complete. Result: {color}{data['result']}{color}")
                return data["message"] == "OK"
            time.sleep(10)

    def _slice_source(self, source: str, offset: list) -> str:
        """Slice the source of the contract, preserving any comments above the first line."""
        offset_start = offset[0]
        top_source = source[:offset_start]
        top_lines = top_source.split("\n")[::-1]
        comment_open = False
        for line in top_lines:
            stripped = line.strip()
            if (
                not stripped
                or stripped.startswith(("//", "/*", "*"))
                or stripped.endswith("*/")
                or comment_open
            ):
                offset_start = offset_start - len(line) - 1
                if stripped.endswith("*/"):
                    comment_open = True
                elif stripped.startswith("/*"):
                    comment_open = False
            else:
                # Stop on the first non-empty, non-comment line
                break
        offset_start = max(0, offset_start)
        return source[offset_start : offset[1]].strip()


class ContractConstructor:
    _dir_color: Final = "bright magenta"

    def __init__(self, parent: "ContractContainer", name: ContractName) -> None:
        self._parent: Final = parent
        try:
            abi = next(i for i in parent.abi if i["type"] == "constructor")
        except Exception:
            abi: ABIConstructor = {"inputs": [], "name": "constructor", "type": "constructor"}
        else:
            abi["name"] = "constructor"
        self.abi: Final = abi
        self._name: Final = name

    @property
    def payable(self) -> bool:
        abi = self.abi
        if "payable" in abi:
            return abi["payable"]
        else:
            return abi["stateMutability"] == "payable"

    def __repr__(self) -> str:
        return f"<{type(self).__name__} '{self._name}.constructor({_inputs(self.abi)})'>"

    def __call__(
        self, *args: Any, publish_source: bool = False, silent: bool = False
    ) -> Union["Contract", TransactionReceiptType]:
        """Deploys a contract.

        Args:
            *args: Constructor arguments. The last argument MUST be a dictionary
                   of transaction values containing at minimum a 'from' key to
                   specify which account to deploy this contract from.

        Returns:
            * Contract instance if the transaction confirms
            * TransactionReceipt if the transaction is pending or reverts"""
        args, tx = _get_tx(None, args)
        if not tx["from"]:
            raise AttributeError(
                "Final argument must be a dict of transaction parameters that "
                "includes a `from` field specifying the address to deploy from"
            )

        return tx["from"].deploy(
            self._parent,
            *args,
            amount=tx["value"],
            gas_limit=tx["gas"],
            gas_price=tx.get("gas_price"),
            max_fee=tx.get("max_fee"),
            priority_fee=tx.get("priority_fee"),
            nonce=tx["nonce"],
            required_confs=tx["required_confs"],
            allow_revert=tx.get("allow_revert"),
            publish_source=publish_source,
            silent=silent,
        )

    @staticmethod
    def _autosuggest(obj: "ContractConstructor") -> list[str]:
        return _contract_method_autosuggest(obj.abi["inputs"], True, obj.payable)

    def encode_input(self, *args: Any) -> str:
        bytecode = self._parent.bytecode
        # find and replace unlinked library pointers in bytecode
        for marker in regex_findall("_{1,}[^_]*_{1,}", bytecode):
            library = marker.strip("_")
            if not self._parent._project[library]:
                raise UndeployedLibrary(
                    f"Contract requires '{library}' library, but it has not been deployed yet"
                )
            address = self._parent._project[library][-1].address[-40:]
            bytecode = bytecode.replace(marker, address)

        abi = self.abi
        data = format_input(abi, args)
        types_list = get_type_strings(abi["inputs"])
        return bytecode + encode_abi(types_list, data).hex()

    def estimate_gas(self, *args: Any) -> int:
        """
        Estimate the gas cost for the deployment.

        Raises VirtualMachineError if the transaction would revert.

        Arguments
        ---------
        *args
            Constructor arguments. The last argument MUST be a dictionary
            of transaction values containing at minimum a 'from' key to
            specify which account to deploy this contract from.

        Returns
        -------
        int
            Estimated gas value in wei.
        """
        args, tx = _get_tx(None, args)
        if not tx["from"]:
            raise AttributeError(
                "Final argument must be a dict of transaction parameters that "
                "includes a `from` field specifying the sender of the transaction"
            )

        return tx["from"].estimate_gas(amount=tx["value"], data=self.encode_input(*args))


class InterfaceContainer:
    """
    Container class that provides access to interfaces within a project.
    """

    def __init__(self, project: Any) -> None:
        self._project = project

        # automatically populate with interfaces in `data/interfaces`
        # overwritten if a project contains an interface with the same name
        for path in BROWNIE_FOLDER.glob("data/interfaces/*.json"):
            with path.open() as fp:
                abi = ujson_load(fp)
            self._add(path.stem, abi)

    def _add(self, name: ContractName, abi: list[ABIElement]) -> None:
        constructor = InterfaceConstructor(name, abi)
        setattr(self, name, constructor)


class InterfaceConstructor:
    """
    Constructor used to create Contract objects from a project interface.
    """

    def __init__(self, name: ContractName, abi: list[ABIElement]) -> None:
        self._name: Final = name
        self.abi: Final = abi
        self.selectors: Final[dict[Selector, FunctionName]] = {
            build_function_selector(i): FunctionName(i["name"])
            for i in abi
            if i["type"] == "function"
        }

    def __call__(self, address: str, owner: AccountsType | None = None) -> "Contract":
        return Contract.from_abi(self._name, address, self.abi, owner, persist=False)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} '{self._name}'>"

    def decode_input(self, calldata: str | bytes) -> tuple[str, Any]:
        """
        Decode input calldata for this contract.

        Arguments
        ---------
        calldata : str | bytes
            Calldata for a call to this contract

        Returns
        -------
        str
            Signature of the function that was called
        Any
            Decoded input arguments
        """
        if not isinstance(calldata, HexBytes):
            calldata = HexBytes(calldata)

        fn_selector = hexbytes_to_hexstring(calldata[:4])

        abi: ABIFunction | None = None
        for _abi in self.abi:
            if _abi["type"] == "function" and build_function_selector(_abi) == fn_selector:
                abi = _abi
                break

        if abi is None:
            raise ValueError("Four byte selector does not match the ABI for this contract")

        function_sig = build_function_signature(abi)

        types_list = get_type_strings(abi["inputs"])
        result = decode_abi(types_list, calldata[4:])
        input_args = format_input(abi, result)

        return function_sig, input_args


class _DeployedContractBase(_ContractBase):
    """Methods for interacting with a deployed contract.

    Each public contract method is available as a ContractCall or ContractTx
    instance, created when this class is instantiated.

    Attributes:
        bytecode: Bytecode of the deployed contract, including constructor args.
        tx: TransactionReceipt of the of the tx that deployed the contract."""

    _reverted = False
    _initialized = False

    def __init__(
        self,
        address: HexAddress,
        owner: AccountsType | None = None,
        tx: TransactionReceiptType = None,
    ) -> None:
        address = _resolve_address(address)
        self.bytecode: Final[HexStr] = (  # type: ignore [assignment]
            # removeprefix is used for compatibility with both hexbytes<1 and >=1
            self._build.get("deployedBytecode", None)
            or web3.eth.get_code(address).hex().removeprefix("0x")
        )
        if not self.bytecode:
            raise ContractNotFound(f"No contract deployed at {address}")
        self._owner: Final = owner
        self.tx: Final = tx
        self.address: Final = address
        self.events = ContractEvents(self)
        _add_deployment_topics(address, self.abi)

        fn_abis = [abi for abi in self.abi if abi["type"] == "function"]
        fn_names = [abi["name"] for abi in fn_abis]

        contract_name = self._name
        contract_natspec: dict = self._build.get("natspec") or {}
        methods_natspec: dict = contract_natspec.get("methods") or {}
        for abi, abi_name in zip(fn_abis, fn_names):
            name = f"{contract_name}.{abi_name}"
            sig = build_function_signature(abi)
            natspec = methods_natspec.get(sig, {})

            if fn_names.count(abi_name) == 1:
                fn = _get_method_object(address, abi, name, owner, natspec)
                self._check_and_set(abi_name, fn)
                continue

            # special logic to handle function overloading
            if not hasattr(self, abi_name):
                overloaded = OverloadedMethod(address, name, owner)
                self._check_and_set(abi_name, overloaded)
            getattr(self, abi_name)._add_fn(abi, natspec)

        self._initialized = True

    def _check_and_set(self, name: str, obj: Any) -> None:
        if name == "balance":
            warnings.warn(
                f"'{self._name}' defines a 'balance' function, "
                f"'{self._name}.balance' is available as {self._name}.wei_balance",
                BrownieEnvironmentWarning,
            )
            setattr(self, "wei_balance", self.balance)
        elif hasattr(self, name):
            warnings.warn(
                "Namespace collision between contract function and "
                f"brownie `Contract` class member: '{self._name}.{name}'\n"
                f"The {name} function will not be available when interacting with {self._name}",
                BrownieEnvironmentWarning,
            )
            return
        setattr(self, name, obj)

    def __hash__(self) -> int:
        return hash(f"{self._name}{self.address}{self._project}")

    def __str__(self) -> str:
        return self.address

    def __repr__(self) -> str:
        alias = self._build.get("alias")
        if alias:
            return f"<'{alias}' Contract '{self.address}'>"
        return f"<{self._name} Contract '{self.address}'>"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, _DeployedContractBase):
            return self.address == other.address and self.bytecode == other.bytecode
        if isinstance(other, str):
            try:
                address = _resolve_address(other)
                return address == self.address
            except ValueError:
                return False
        return super().__eq__(other)

    def __getattribute__(self, name: str) -> AnyContractMethod:
        if super().__getattribute__("_reverted"):
            raise ContractNotFound("This contract no longer exists.")
        try:
            return super().__getattribute__(name)
        except AttributeError:
            raise AttributeError(f"Contract '{self._name}' object has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        if self._initialized and isinstance(getattr(self, name, None), _ContractMethod):
            raise AttributeError(
                f"{self._name}.{name} is a contract function, it cannot be assigned to"
            )
        super().__setattr__(name, value)

    def get_method_object(self, calldata: str) -> Optional["_ContractMethod"]:
        """
        Given a calldata hex string, returns a `ContractMethod` object.
        """
        sig = calldata[:10].lower()
        if sig not in self.selectors:
            return None
        fn = getattr(self, self.selectors[sig], None)
        if isinstance(fn, OverloadedMethod):
            return next((v for v in fn.methods.values() if v.signature == sig), None)
        return fn

    def balance(self) -> Wei:
        """Returns the current ether balance of the contract, in wei."""
        balance = web3.eth.get_balance(self.address)
        return Wei(balance)

    def _deployment_path(self) -> Path | None:
        if not self._project._path or (
            CONFIG.network_type != "live" and not CONFIG.settings["dev_deployment_artifacts"]
        ):
            return None

        chainid = CONFIG.active_network["chainid"] if CONFIG.network_type == "live" else "dev"
        path = self._project._build_path.joinpath(f"deployments/{chainid}")
        path.mkdir(exist_ok=True)
        return path.joinpath(f"{self.address}.json")

    def _save_deployment(self) -> None:
        path = self._deployment_path()
        chainid = CONFIG.active_network["chainid"] if CONFIG.network_type == "live" else "dev"
        deployment_build = self._build.copy()

        deployment_build["deployment"] = {
            "address": self.address,
            "chainid": chainid,
            "blockHeight": web3.eth.block_number,
        }
        if path:
            self._project._add_to_deployment_map(self)
            if not path.exists():
                with path.open("w") as fp:
                    ujson_dump(deployment_build, fp)

    def _delete_deployment(self) -> None:
        if path := self._deployment_path():
            self._project._remove_from_deployment_map(self)
            if path.exists():
                path.unlink()


class Contract(_DeployedContractBase):
    """
    Object to interact with a deployed contract outside of a project.
    """

    def __init__(
        self,
        address_or_alias: HexAddress | ContractName,
        *args: Any,
        owner: AccountsType | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Recreate a `Contract` object from the local database.

        The init method is used to access deployments that have already previously
        been stored locally. For new deployments use `from_abi` or `from_etherscan`.

        Arguments
        ---------
        address_or_alias : str
            Address or user-defined alias of the deployment.
        owner : Account, optional
            Contract owner. If set, transactions without a `from` field
            will be performed using this account.
        """
        address_or_alias = address_or_alias.strip()

        if args or kwargs:
            warnings.warn(
                "Initializing `Contract` in this manner is deprecated." " Use `from_abi` instead.",
                DeprecationWarning,
            )
            kwargs["owner"] = owner
            return self._deprecated_init(address_or_alias, *args, **kwargs)

        address = ""
        try:
            address = _resolve_address(address_or_alias)
            build, sources = _get_deployment(address)
        except Exception:
            build, sources = _get_deployment(alias=address_or_alias)
            if build is not None:
                address = build["address"]

        if build is None or sources is None:
            if (
                not address
                or not CONFIG.settings.get("autofetch_sources")
                or not CONFIG.active_network.get("explorer")
            ):
                if not address:
                    raise ValueError(f"Unknown alias: '{address_or_alias}'")
                else:
                    raise ValueError(f"Unknown contract address: '{address}'")
            contract = self.from_explorer(address, owner=owner, silent=True)
            build, sources = contract._build, contract._sources
            address = contract.address

        _ContractBase.__init__(self, None, build, sources)
        _DeployedContractBase.__init__(self, address, owner)

    def _deprecated_init(
        self,
        name: ContractName,
        address: HexAddress | None = None,
        abi: list[ABIElement] | None = None,
        manifest_uri: str | None = None,
        owner: AccountsType | None = None,
    ) -> None:
        if manifest_uri:
            raise ValueError("ethPM functionality removed")

        if not address:
            raise TypeError("Address cannot be None unless creating object from manifest")

        build = {"abi": abi, "contractName": name, "type": "contract"}
        _ContractBase.__init__(self, None, build, {})
        _DeployedContractBase.__init__(self, address, owner, None)

    @classmethod
    def from_abi(
        cls,
        name: ContractName,
        address: HexAddress,
        abi: list[ABIElement],
        owner: AccountsType | None = None,
        persist: bool = True,
    ) -> "Contract":
        """
        Create a new `Contract` object from an ABI.

        Arguments
        ---------
        name : str
            Name of the contract.
        address : str
            Address where the contract is deployed.
        abi : dict
            Contract ABI, given as a dictionary.
        owner : Account, optional
            Contract owner. If set, transactions without a `from` field
            will be performed using this account.
        """
        address = _resolve_address(address)
        build = {
            "abi": abi,
            "address": address,
            "contractName": name,
            "type": "contract",
            # removeprefix is used for compatibility with both hexbytes<1 and >=1
            "deployedBytecode": web3.eth.get_code(address).hex().removeprefix("0x"),
        }

        self = cls.__new__(cls)
        _ContractBase.__init__(self, None, build, {})
        _DeployedContractBase.__init__(self, address, owner, None)
        if persist:
            _add_deployment(self)
        return self

    @classmethod
    def from_explorer(
        cls,
        address: HexAddress,
        as_proxy_for: str | None = None,
        owner: AccountsType | None = None,
        silent: bool = False,
        persist: bool = True,
    ) -> "Contract":
        """
        Create a new `Contract` object with source code queried from a block explorer.

        Arguments
        ---------
        address : str
            Address where the contract is deployed.
        as_proxy_for : str, optional
            Address of the implementation contract, if `address` is a proxy contract.
            The generated object will send transactions to `address`, but use the ABI
            and NatSpec of `as_proxy_for`. This field is only required when the
            block explorer API does not provide an implementation address.
        owner : Account, optional
            Contract owner. If set, transactions without a `from` field will be
            performed using this account.
        """
        address = _resolve_address(address)
        data = _fetch_from_explorer(address, "getsourcecode", silent)
        is_verified = bool(data["result"][0].get("SourceCode"))

        if is_verified:
            abi = ujson_loads(data["result"][0]["ABI"])
            name = data["result"][0]["ContractName"]
        else:
            # if the source is not available, try to fetch only the ABI
            try:
                data_abi = _fetch_from_explorer(address, "getabi", True)
            except ValueError as exc:
                _unverified_addresses.add(address)
                raise exc
            abi = ujson_loads(data_abi["result"].strip())
            name = "UnknownContractName"
            if not silent:
                warnings.warn(
                    f"{address}: Was able to fetch the ABI but not the source code. "
                    "Some functionality will not be available.",
                    BrownieCompilerWarning,
                )

        if as_proxy_for is None:
            # always check for an EIP1967 proxy - https://eips.ethereum.org/EIPS/eip-1967
            implementation_eip1967 = web3.eth.get_storage_at(
                address, int(web3.keccak(text="eip1967.proxy.implementation").hex(), 16) - 1
            )
            # always check for an EIP1822 proxy - https://eips.ethereum.org/EIPS/eip-1822
            implementation_eip1822 = web3.eth.get_storage_at(address, web3.keccak(text="PROXIABLE"))
            if len(implementation_eip1967) > 0 and int(implementation_eip1967.hex(), 16):
                as_proxy_for = _resolve_address(implementation_eip1967[-20:])
            elif len(implementation_eip1822) > 0 and int(implementation_eip1822.hex(), 16):
                as_proxy_for = _resolve_address(implementation_eip1822[-20:])
            elif data["result"][0].get("Implementation"):
                # for other proxy patterns, we only check if etherscan indicates
                # the contract is a proxy. otherwise we could have a false positive
                # if there is an `implementation` method on a regular contract.
                try:
                    # first try to call `implementation` per EIP897
                    # https://eips.ethereum.org/EIPS/eip-897
                    contract = cls.from_abi(name, address, abi)
                    as_proxy_for = contract.implementation.call()
                except Exception:
                    # if that fails, fall back to the address provided by etherscan
                    as_proxy_for = _resolve_address(data["result"][0]["Implementation"])

        if as_proxy_for == address:
            as_proxy_for = None

        # if this is a proxy, fetch information for the implementation contract
        if as_proxy_for is not None:
            implementation_contract = Contract.from_explorer(as_proxy_for)
            abi = implementation_contract._build["abi"]

        if not is_verified:
            return cls.from_abi(name, address, abi, owner)

        compiler_str = data["result"][0]["CompilerVersion"]
        if compiler_str.startswith("vyper:"):
            try:
                version = to_vyper_version(compiler_str[6:])
                is_compilable = version in get_installable_vyper_versions()
            except Exception:
                is_compilable = False
        else:
            try:
                version = cls.get_solc_version(compiler_str, address)

                is_compilable = (
                    version >= Version("0.4.22")
                    and version
                    in solcx.get_installable_solc_versions() + solcx.get_installed_solc_versions()
                )
            except Exception:
                is_compilable = False

        if not is_compilable:
            if not silent:
                warnings.warn(
                    f"{address}: target compiler '{compiler_str}' cannot be installed or is not "
                    "supported by Brownie. Some debugging functionality will not be available.",
                    BrownieCompilerWarning,
                )
            return cls.from_abi(name, address, abi, owner)
        elif data["result"][0]["OptimizationUsed"] in ("true", "false"):
            if not silent:
                warnings.warn(
                    "Blockscout explorer API has limited support by Brownie. "  # noqa
                    "Some debugging functionality will not be available.",
                    BrownieCompilerWarning,
                )
            return cls.from_abi(name, address, abi, owner)

        optimizer = {
            "enabled": bool(int(data["result"][0]["OptimizationUsed"])),
            "runs": int(data["result"][0]["Runs"]),
        }
        evm_version = data["result"][0].get("EVMVersion", "Default")
        if evm_version == "Default":
            evm_version = None

        source_str = "\n".join(data["result"][0]["SourceCode"].splitlines())
        try:
            if source_str.startswith("{{"):
                # source was verified using compiler standard JSON
                input_json = ujson_loads(source_str[1:-1])
                sources = {k: v["content"] for k, v in input_json["sources"].items()}
                evm_version = input_json["settings"].get("evmVersion", evm_version)
                remappings = input_json["settings"].get("remappings", [])

                compiler.set_solc_version(str(version))
                input_json.update(
                    compiler.generate_input_json(
                        sources, optimizer=optimizer, evm_version=evm_version, remappings=remappings
                    )
                )
                output_json = compiler.compile_from_input_json(input_json)
                build_json = compiler.generate_build_json(input_json, output_json)
            else:
                if source_str.startswith("{"):
                    # source was submitted as multiple files
                    sources = {k: v["content"] for k, v in ujson_loads(source_str).items()}
                else:
                    # source was submitted as a single file
                    if compiler_str.startswith("vyper"):
                        path_str = f"{name}.vy"
                    else:
                        path_str = f"{name}-flattened.sol"
                    sources = {path_str: source_str}

                build_json = compiler.compile_and_format(
                    sources,
                    solc_version=str(version),
                    vyper_version=str(version),
                    optimizer=optimizer,
                    evm_version=evm_version,
                )
        except Exception as e:
            if not silent:
                warnings.warn(
                    f"{address}: Compilation failed due to {type(e).__name__}. Falling back to ABI,"
                    " some functionality will not be available.",
                    BrownieCompilerWarning,
                )
            return cls.from_abi(name, address, abi, owner)

        build_json = build_json[name]
        if as_proxy_for is not None:
            build_json.update(abi=abi, natspec=implementation_contract._build.get("natspec"))

        if not _verify_deployed_code(
            address, build_json["deployedBytecode"], build_json["language"]
        ):
            if not silent:
                warnings.warn(
                    f"{address}: Locally compiled and on-chain bytecode do not match!",
                    BrownieCompilerWarning,
                )
            del build_json["pcMap"]

        self = cls.__new__(cls)
        _ContractBase.__init__(self, None, build_json, sources)
        _DeployedContractBase.__init__(self, address, owner)
        if persist:
            _add_deployment(self)
        return self

    @classmethod
    def get_solc_version(cls, compiler_str: str, address: str) -> Version:
        """
        Return the solc compiler version either from the passed compiler string
        or try to find the latest available patch semver compiler version.

        Arguments
        ---------
        compiler_str: str
            The compiler string passed from the contract metadata.
        address: str
            The contract address to check for.
        """
        version = Version(compiler_str.lstrip("v")).truncate()

        compiler_config = _load_project_compiler_config(Path(os.getcwd()))
        solc_config = compiler_config["solc"]
        if "use_latest_patch" in solc_config:
            use_latest_patch = solc_config["use_latest_patch"]
            needs_patch_version = False
            if isinstance(use_latest_patch, bool):
                needs_patch_version = use_latest_patch
            elif isinstance(use_latest_patch, list):
                needs_patch_version = address in use_latest_patch

            if needs_patch_version:
                versions = [Version(str(i)) for i in solcx.get_installable_solc_versions()]
                for v in filter(lambda x: x < version.next_minor(), versions):
                    if v > version:
                        version = v

        return version

    @classmethod
    def remove_deployment(
        cls,
        address: ChecksumAddress | None = None,
        alias: ContractName | None = None,
    ) -> tuple[dict | None, dict | None]:
        """
        Removes this contract from the internal deployments db
        with the passed address or alias.

        Arguments
        ---------
        address: str | None
            An address to apply
        alias: str | None
            An alias to apply
        """
        return _remove_deployment(address, alias)

    def set_alias(self, alias: str | None, persist: bool = True) -> None:
        """
        Apply a unique alias this object. The alias can be used to restore the
        object in future sessions.

        Arguments
        ---------
        alias: str | None
            An alias to apply. If `None`, any existing alias is removed.
        """
        if "chainid" not in CONFIG.active_network:
            raise ValueError("Cannot set aliases in a development environment")

        if alias is not None:
            if "." in alias or alias.lower().startswith("0x"):
                raise ValueError("Invalid alias")
            build, _ = _get_deployment(alias=alias)
            if build is not None:
                if build["address"] != self.address:
                    raise ValueError("Alias is already in use on another contract")
                return

        if persist:
            _add_deployment(self, alias)
        self._build["alias"] = alias

    @property
    def alias(self) -> str | None:
        return self._build.get("alias")


class ProjectContract(_DeployedContractBase):
    """Methods for interacting with a deployed contract as part of a Brownie project."""

    def __init__(
        self,
        project: Any,
        build: ContractBuildJson,
        address: ChecksumAddress,
        owner: AccountsType | None = None,
        tx: TransactionReceiptType = None,
    ) -> None:
        _ContractBase.__init__(self, project, build, project._sources)
        _DeployedContractBase.__init__(self, address, owner, tx)


class ContractEvents(_ContractEvents):
    def __init__(self, contract: _DeployedContractBase):
        self.linked_contract = contract

        # Ignoring type since ChecksumAddress type is an alias for string
        _ContractEvents.__init__(self, contract.abi, web3, contract.address)

    def subscribe(
        self, event_name: str, callback: Callable[[AttributeDict], None], delay: float = 2.0
    ) -> None:
        """
        Subscribe to event with a name matching 'event_name', calling the 'callback'
        function on new occurrence giving as parameter the event log receipt.

        Args:
            event_name (str): Name of the event to subscribe to.
            callback (Callable[[AttributeDict], None]): Function called whenever an event occurs.
            delay (float, optional): Delay between each check for new events. Defaults to 2.0.
        """
        target_event: ContractEvent = self.__getitem__(event_name)
        event_watcher.add_event_callback(event=target_event, callback=callback, delay=delay)

    def get_sequence(
        self, from_block: int, to_block: int = None, event_type: ContractEvent | str = None
    ) -> list[AttributeDict] | AttributeDict:
        """Returns the logs of events of type 'event_type' that occurred between the
        blocks 'from_block' and 'to_block'. If 'event_type' is not specified,
        it retrieves the occurrences of all events in the contract.

        Args:
            from_block (int): The block from which to search for events that have occurred.
            to_block (int, optional): The block on which to stop searching for events.
            if not specified, it is set to the most recently mined block (web3.eth.block_number).
            Defaults to None.
            event_type (ContractEvent, str, optional): Type or name of the event to be searched
            between the specified blocks. Defaults to None.

        Returns:
            if 'event_type' is specified:
                [list]: List of events of type 'event_type' that occurred between
                'from_block' and 'to_block'.
            else:
                event_logbook [dict]: Dictionary of events of the contract that occurred
                between 'from_block' and 'to_block'.
        """
        if to_block is None or to_block > web3.eth.block_number:
            to_block = web3.eth.block_number

        # Returns event sequence for the specified event
        if event_type is not None:
            if isinstance(event_type, str):
                # If 'event_type' is a string, search for an event with a name matching it.
                event_type: ContractEvent = self.__getitem__(event_type)
            return self._retrieve_contract_events(event_type, from_block, to_block)

        return AttributeDict(
            (event.event_name, self._retrieve_contract_events(event, from_block, to_block))
            for event in ContractEvents.__iter__(self)
        )

    def listen(self, event_name: str, timeout: float = 0) -> Coroutine:
        """
        Creates a listening Coroutine object ending whenever an event matching
        'event_name' occurs. If timeout is superior to zero and no event matching
        'event_name' has occurred, the Coroutine ends when the timeout is reached.

        The Coroutine return value is an AttributeDict filled with the following fields :
            - 'event_data' (AttributeDict): The event log receipt that was caught.
            - 'timed_out' (bool): False if the event did not timeout, else True

        If the 'timeout' parameter is not passed or is inferior or equal to 0,
        the Coroutine listens indefinitely.

        Args:
            event_name (str): Name of the event to be listened to.
            timeout (float, optional): Timeout value in seconds. Defaults to 0.

        Returns:
            Coroutine: Awaitable object listening for the event matching 'event_name'.
        """
        _triggered: bool = False
        _received_data: AttributeDict | None = None

        def _event_callback(event_data: AttributeDict) -> None:
            """
            Fills the nonlocal variable '_received_data' with the received
            argument 'event_data' and sets the nonlocal '_triggered' variable to True
            """
            nonlocal _triggered, _received_data
            _received_data = event_data
            _triggered = True

        _listener_end_time = time.time() + timeout

        async def _listening_task(is_timeout: bool, end_time: float) -> AttributeDict:
            """Generates and returns a coroutine listening for an event"""
            nonlocal _triggered, _received_data
            timed_out: bool = False

            while not _triggered:
                if is_timeout and end_time <= time.time():
                    timed_out = True
                    break
                await asyncio.sleep(0.05)
            return AttributeDict({"event_data": _received_data, "timed_out": timed_out})

        target_event: ContractEvent = self.__getitem__(event_name)
        event_watcher.add_event_callback(
            event=target_event, callback=_event_callback, delay=0.2, repeat=False
        )
        return _listening_task(timeout > 0, _listener_end_time)

    @combomethod
    def _retrieve_contract_events(
        self, event_type: ContractEvent, from_block: int = None, to_block: int = None
    ) -> list[LogReceipt]:
        """
        Retrieves all log receipts from 'event_type' between 'from_block' and 'to_block' blocks
        """
        if to_block is None:
            to_block = web3.eth.block_number
        if from_block is None and isinstance(to_block, int):
            from_block = to_block - 10

        event_filter: filters.LogFilter = event_type.create_filter(
            fromBlock=from_block, toBlock=to_block
        )
        return event_filter.get_all_entries()


class OverloadedMethod:
    def __init__(self, address: ChecksumAddress, name: str, owner: AccountsType | None):
        self._address: Final = address
        self._name: Final = name
        self._owner: Final = owner
        self.methods: Final[dict[Any, ContractCall | ContractTx]] = {}
        self.natspec: Final[dict[str, Any]] = {}

    def _add_fn(self, abi: ABIFunction, natspec: dict[str, Any]) -> None:
        fn = _get_method_object(self._address, abi, self._name, self._owner, natspec)
        key = tuple(i["type"].replace("256", "") for i in abi["inputs"])
        self.methods[key] = fn
        self.natspec.update(natspec)

    def _get_fn_from_args(self, args: tuple) -> "_ContractMethod":
        input_length = len(args)
        if args and isinstance(args[-1], dict):
            input_length -= 1
        keys = [i for i in self.methods if len(i) == input_length]
        if not keys:
            raise ValueError("No function matching the given number of arguments")
        if len(keys) > 1:
            raise ValueError(
                f"Contract has more than one function '{self._name}' requiring "
                f"{input_length} arguments. You must explicitly declare which function "
                f"you are calling, e.g. {self._name}['{','.join(keys[0])}'](*args)"
            )
        return self.methods[keys[0]]

    def __getitem__(self, key: tuple[str, ...] | str) -> Union["ContractCall", "ContractTx"]:
        if isinstance(key, str):
            key = (i.strip() for i in key.split(","))

        key = tuple(i.replace("256", "") for i in key)
        return self.methods[key]

    def __repr__(self) -> str:
        return f"<OverloadedMethod '{self._name}'>"

    def __len__(self) -> int:
        return len(self.methods)

    def __call__(
        self, *args: Any, block_identifier: int | str | bytes = None, override: dict = None
    ) -> Any:
        fn = self._get_fn_from_args(args)
        kwargs = {"block_identifier": block_identifier, "override": override}
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return fn(*args, **kwargs)

    def call(
        self, *args: Any, block_identifier: int | str | bytes = None, override: dict = None
    ) -> Any:
        """
        Call the contract method without broadcasting a transaction.

        The specific function called is chosen based on the number of
        arguments given. If more than one function exists with this number
        of arguments, a `ValueError` is raised.

        Arguments
        ---------
        *args
            Contract method inputs. You can optionally provide a
            dictionary of transaction properties as the last arg.
        block_identifier : int | str | bytes, optional
            A block number or hash that the call is executed at. If not given, the
            latest block used. Raises `ValueError` if this value is too far in the
            past and you are not using an archival node.
        override : dict, optional
            A mapping from addresses to balance, nonce, code, state, stateDiff
            overrides for the context of the call.

        Returns
        -------
            Contract method return value(s).
        """
        fn = self._get_fn_from_args(args)
        return fn.call(*args, block_identifier=block_identifier, override=override)

    def transact(self, *args: Any) -> TransactionReceiptType:
        """
        Broadcast a transaction that calls this contract method.

        The specific function called is chosen based on the number of
        arguments given. If more than one function exists with this number
        of arguments, a `ValueError` is raised.

        Arguments
        ---------
        *args
            Contract method inputs. You can optionally provide a
            dictionary of transaction properties as the last arg.

        Returns
        -------
        TransactionReceipt
            Object representing the broadcasted transaction.
        """
        fn = self._get_fn_from_args(args)
        return fn.transact(*args)

    def encode_input(self, *args: Any) -> Any:
        """
        Generate encoded ABI data to call the method with the given arguments.

        Arguments
        ---------
        *args
            Contract method inputs

        Returns
        -------
        str
            Hexstring of encoded ABI data
        """
        fn = self._get_fn_from_args(args)
        return fn.encode_input(*args)

    def decode_input(self, hexstr: str) -> list:
        """
        Decode input call data for this method.

        Arguments
        ---------
        hexstr : str
            Hexstring of input call data

        Returns
        -------
        Decoded values
        """
        selector = hexbytes_to_hexstring(HexBytes(hexstr)[:4])
        for fn in self.methods.values():
            if fn == selector:
                return fn.decode_input(hexstr)
        raise ValueError(
            "Data cannot be decoded using any input signatures of functions of this name"
        )

    def decode_output(self, hexstr: str) -> tuple:
        """
        Decode hexstring data returned by this method.

        Arguments
        ---------
        hexstr : str
            Hexstring of returned call data

        Returns
        -------
        Decoded values
        """
        for fn in self.methods.values():
            try:
                return fn.decode_output(hexstr)
            except Exception:
                pass
        raise ValueError(
            "Data cannot be decoded using any output signatures of functions of this name"
        )

    def info(self) -> None:
        """
        Display NatSpec documentation for this method.
        """
        fn_sigs = (f"{fn.abi['name']}({_inputs(fn.abi)})" for fn in self.methods.values())
        for sig in sorted(fn_sigs, key=lambda k: len(k)):
            print(sig)
        _print_natspec(self.natspec)


class _ContractMethod:
    _dir_color: Final = "bright magenta"

    def __init__(
        self,
        address: ChecksumAddress,
        abi: ABIFunction,
        name: str,
        owner: AccountsType | None,
        natspec: dict[str, Any] | None = None,
    ) -> None:
        self._address: Final = address
        self._name: Final = name
        self.abi: Final = abi
        self._owner: Final = owner
        self.signature: Final = build_function_selector(abi)
        self._input_sig: Final = build_function_signature(abi)
        self.natspec: Final[dict[str, Any]] = natspec or {}

    def __repr__(self) -> str:
        pay = "payable " if self.payable else ""
        return f"<{type(self).__name__} {pay}'{self.abi['name']}({_inputs(self.abi)})'>"

    @property
    def payable(self) -> bool:
        abi = self.abi
        if "payable" in abi:
            return abi["payable"]
        else:
            return abi["stateMutability"] == "payable"

    @staticmethod
    def _autosuggest(obj: "_ContractMethod") -> list[str]:
        # this is a staticmethod to be compatible with `_call_suggest` and `_transact_suggest`
        return _contract_method_autosuggest(
            obj.abi["inputs"], isinstance(obj, ContractTx), obj.payable
        )

    def info(self) -> None:
        """
        Display NatSpec documentation for this method.
        """
        abi = self.abi
        print(f"{abi['name']}({_inputs(abi)})")
        _print_natspec(self.natspec)

    def call(
        self, *args: Any, block_identifier: int | str | bytes = None, override: dict = None
    ) -> Any:
        """
        Call the contract method without broadcasting a transaction.

        Arguments
        ---------
        *args
            Contract method inputs. You can optionally provide a
            dictionary of transaction properties as the last arg.
        block_identifier : int | str | bytes, optional
            A block number or hash that the call is executed at. If not given, the
            latest block used. Raises `ValueError` if this value is too far in the
            past and you are not using an archival node.
        override : dict, optional
            A mapping from addresses to balance, nonce, code, state, stateDiff
            overrides for the context of the call.

        Returns
        -------
            Contract method return value(s).
        """

        args, tx = _get_tx(self._owner, args)
        if tx["from"]:
            tx["from"] = str(tx["from"])
        del tx["required_confs"]
        tx.update({"to": self._address, "data": self.encode_input(*args)})

        try:
            data = web3.eth.call({k: v for k, v in tx.items() if v}, block_identifier, override)
        except ValueError as e:
            raise VirtualMachineError(e) from None

        if self.abi["outputs"] and not data:
            raise ValueError("No data was returned - the call likely reverted")
        try:
            return self.decode_output(data)
        except Exception:
            raise ValueError(f"Call reverted: {decode_typed_error(data)}") from None

    def transact(self, *args: Any, silent: bool = False) -> TransactionReceiptType:
        """
        Broadcast a transaction that calls this contract method.

        Arguments
        ---------
        *args
            Contract method inputs. You can optionally provide a
            dictionary of transaction properties as the last arg.

        Returns
        -------
        TransactionReceipt
            Object representing the broadcasted transaction.
        """

        args, tx = _get_tx(self._owner, args)
        if not tx["from"]:
            raise AttributeError(
                "Final argument must be a dict of transaction parameters that "
                "includes a `from` field specifying the sender of the transaction"
            )

        return tx["from"].transfer(
            self._address,
            tx["value"],
            gas_limit=tx["gas"],
            gas_buffer=tx.get("gas_buffer"),
            gas_price=tx.get("gas_price"),
            max_fee=tx.get("max_fee"),
            priority_fee=tx.get("priority_fee"),
            nonce=tx["nonce"],
            required_confs=tx["required_confs"],
            data=self.encode_input(*args),
            allow_revert=tx["allow_revert"],
            silent=silent,
        )

    def decode_input(self, hexstr: str) -> list:
        """
        Decode input call data for this method.

        Arguments
        ---------
        hexstr : str
            Hexstring of input call data

        Returns
        -------
        Decoded values
        """
        abi = self.abi
        types_list = get_type_strings(abi["inputs"])
        result = decode_abi(types_list, HexBytes(hexstr)[4:])
        return format_input(abi, result)

    def encode_input(self, *args: Any) -> str:
        """
        Generate encoded ABI data to call the method with the given arguments.

        Arguments
        ---------
        *args
            Contract method inputs

        Returns
        -------
        str
            Hexstring of encoded ABI data
        """
        abi = self.abi
        data = format_input(abi, args)
        types_list = get_type_strings(abi["inputs"])
        return self.signature + encode_abi(types_list, data).hex()

    def decode_output(self, hexstr: str) -> tuple:
        """
        Decode hexstring data returned by this method.

        Arguments
        ---------
        hexstr : str
            Hexstring of returned call data

        Returns
        -------
        Decoded values
        """
        abi = self.abi
        types_list = get_type_strings(abi["outputs"])
        result = decode_abi(types_list, HexBytes(hexstr))
        result = format_output(abi, result)
        if len(result) == 1:
            result = result[0]
        return result

    def estimate_gas(self, *args: Any) -> int:
        """
        Estimate the gas cost for a transaction.

        Raises VirtualMachineError if the transaction would revert.

        Arguments
        ---------
        *args
            Contract method inputs

        Returns
        -------
        int
            Estimated gas value in wei.
        """
        args, tx = _get_tx(self._owner, args)
        if not tx["from"]:
            raise AttributeError(
                "Final argument must be a dict of transaction parameters that "
                "includes a `from` field specifying the sender of the transaction"
            )

        return tx["from"].estimate_gas(
            to=self._address,
            amount=tx["value"],
            data=self.encode_input(*args),
        )


class ContractTx(_ContractMethod):
    """
    A public payable or non-payable contract method.

    Attributes
    ----------
    abi : dict
        Contract ABI specific to this method.
    signature : str
        Bytes4 method signature.
    """

    def __call__(self, *args: Any, silent: bool = False) -> TransactionReceiptType:
        """
        Broadcast a transaction that calls this contract method.

        Arguments
        ---------
        *args
            Contract method inputs. You can optionally provide a
            dictionary of transaction properties as the last arg.

        Returns
        -------
        TransactionReceipt
            Object representing the broadcasted transaction.
        """

        return self.transact(*args, silent=silent)


class ContractCall(_ContractMethod):
    """
    A public view or pure contract method.

    Attributes
    ----------
    abi : dict
        Contract ABI specific to this method.
    signature : str
        Bytes4 method signature.
    """

    def __call__(
        self, *args: Any, block_identifier: int | str | bytes = None, override: dict = None
    ) -> Any:
        """
        Call the contract method without broadcasting a transaction.

        Arguments
        ---------
        args
            Contract method inputs. You can optionally provide a
            dictionary of transaction properties as the last arg.
        block_identifier : int | str | bytes, optional
            A block number or hash that the call is executed at. If not given, the
            latest block used. Raises `ValueError` if this value is too far in the
            past and you are not using an archival node.
        override : dict, optional
            A mapping from addresses to balance, nonce, code, state, stateDiff
            overrides for the context of the call.

        Returns
        -------
            Contract method return value(s).
        """

        if not CONFIG.argv["always_transact"] or block_identifier is not None:
            return self.call(*args, block_identifier=block_identifier, override=override)

        args, tx = _get_tx(self._owner, args)
        tx.update({"gas_price": 0, "from": self._owner or accounts[0]})
        pc, revert_msg = None, None

        try:
            self.transact(*args, tx)
            chain.undo()
        except VirtualMachineError as exc:
            pc, revert_msg = exc.pc, exc.revert_msg
            chain.undo()
        except Exception:
            pass

        try:
            return self.call(*args)
        except VirtualMachineError as exc:
            if pc == exc.pc and revert_msg and exc.revert_msg is None:
                # in case we miss a dev revert string
                exc.revert_msg = revert_msg
            raise exc


def _get_tx(owner: AccountsType | None, args: tuple) -> tuple:
    # set / remove default sender
    if owner is None:
        owner = accounts.default
    default_owner = CONFIG.active_network["settings"]["default_contract_owner"]
    if CONFIG.mode == "test" and default_owner is False:
        owner = None

    # separate contract inputs from tx dict and set default tx values
    tx = {
        "from": owner,
        "value": 0,
        "gas": None,
        "gas_buffer": None,
        "nonce": None,
        "required_confs": 1,
        "allow_revert": None,
    }
    if args and isinstance(args[-1], dict):
        tx.update(args[-1])
        args = args[:-1]
        # key substitution to provide compatibility with web3.py
        for key, target in [("amount", "value"), ("gas_limit", "gas"), ("gas_price", "gasPrice")]:
            if key in tx:
                tx[target] = tx[key]

    # enable the magic of ganache's `evm_unlockUnknownAccount`
    if isinstance(tx["from"], str):
        tx["from"] = accounts.at(tx["from"], force=True)
    elif isinstance(tx["from"], _DeployedContractBase):
        tx["from"] = accounts.at(tx["from"].address, force=True)

    return args, tx


def _get_method_object(
    address: ChecksumAddress,
    abi: ABIFunction,
    name: str,
    owner: AccountsType | None,
    natspec: dict[str, Any],
) -> ContractCall | ContractTx:
    if "constant" in abi:
        constant = abi["constant"]
    else:
        constant = abi["stateMutability"] in ("view", "pure")

    if constant:
        return ContractCall(address, abi, name, owner, natspec)
    return ContractTx(address, abi, name, owner, natspec)


_fixed168x10: Final = {"fixed168x10": "decimal"}


def _inputs(abi: ABIFunction | ABIConstructor) -> str:
    abi_inputs = abi["inputs"]
    types_list = get_type_strings(abi_inputs, _fixed168x10)
    return ", ".join(
        f"{types}{bright_blue}{f' {name}' if name else ''}{color}"
        for name, types in zip((i["name"] for i in abi_inputs), types_list)
    )


def _verify_deployed_code(
    address: ChecksumAddress, expected_bytecode: HexStr, language: Language
) -> bool:
    # removeprefix is used for compatibility with both hexbytes<1 and >=1
    actual_bytecode = web3.eth.get_code(address).hex().removeprefix("0x")
    expected_bytecode = expected_bytecode.removeprefix("0x")

    if expected_bytecode.startswith("730000000000000000000000000000000000000000"):
        # special case for Solidity libraries
        return (
            actual_bytecode.startswith(f"73{address[2:].lower()}")
            and actual_bytecode[42:] == expected_bytecode[42:]
        )

    if "_" in expected_bytecode:
        for marker in regex_findall("_{1,}[^_]*_{1,}", expected_bytecode):
            idx = expected_bytecode.index(marker)
            actual_bytecode = actual_bytecode[:idx] + actual_bytecode[idx + 40 :]
            expected_bytecode = expected_bytecode[:idx] + expected_bytecode[idx + 40 :]

    if language == "Solidity":
        # do not include metadata in comparison
        idx = -(int(actual_bytecode[-4:], 16) + 2) * 2
        actual_bytecode = actual_bytecode[:idx]
        idx = -(int(expected_bytecode[-4:], 16) + 2) * 2
        expected_bytecode = expected_bytecode[:idx]

    if language == "Vyper":
        # don't check immutables section
        # TODO actually grab data section length from layout.
        return actual_bytecode.startswith(expected_bytecode)

    return actual_bytecode == expected_bytecode


def _print_natspec(natspec: dict[str, Any]) -> None:
    wrapper = TextWrapper(initial_indent=f"  {bright_magenta}")
    for key in ("title", "notice", "author", "details"):
        if key in natspec:
            wrapper.subsequent_indent = " " * (len(key) + 4)
            print(wrapper.fill(f"@{key} {color}{natspec[key]}"))

    for key, value in natspec.get("params", {}).items():
        wrapper.subsequent_indent = " " * 9
        print(wrapper.fill(f"@param {bright_blue}{key}{color} {value}"))

    if "return" in natspec:
        wrapper.subsequent_indent = " " * 10
        print(wrapper.fill(f"@return {color}{natspec['return']}"))

    for key in sorted(natspec.get("returns", [])):
        wrapper.subsequent_indent = " " * 10
        print(wrapper.fill(f"@return {color}{natspec['returns'][key]}"))

    print()


def _fetch_from_explorer(address: ChecksumAddress, action: str, silent: bool) -> dict[str, Any]:
    if address in _unverified_addresses:
        raise ValueError(f"Source for {address} has not been verified")

    # removeprefix is used for compatibility with both hexbytes<1 and >=1
    code = web3.eth.get_code(address).hex().removeprefix("0x")
    # EIP-1167: Minimal Proxy Contract
    if code[:20] == "363d3d373d3d3d363d73" and code[60:] == "5af43d82803e903d91602b57fd5bf3":
        address = _resolve_address(code[20:60])
    # Vyper <0.2.9 `create_forwarder_to`
    elif (
        code[:30] == "366000600037611000600036600073"
        and code[70:] == "5af4602c57600080fd5b6110006000f3"
    ):
        address = _resolve_address(code[30:70])
    # 0xSplits Clones
    elif (
        code[:120]
        == "36603057343d52307f830d2d700a97af574b186c80d40429385d24241565b08a7c559ba283a964d9b160203da23d3df35b3d3d3d3d363d3d37363d73"  # noqa e501
        and code[160:] == "5af43d3d93803e605b57fd5bf3"
    ):
        address = _resolve_address(code[120:160])

    params: dict[str, Any] = {
        "module": "contract",
        "action": action,
        "address": address,
        "chainid": web3.chain_id,
    }
    env_key = os.getenv("ETHERSCAN_TOKEN")
    if env_key is not None:
        params["apiKey"] = env_key
    elif not silent:
        warnings.warn(
            "No ETHERSCAN_API token set. You may experience issues with rate limiting. "
            "Visit https://etherscan.io/register to obtain a token, and then store it "
            "as the environment variable $ETHERSCAN_TOKEN",
            BrownieEnvironmentWarning,
        )
    if not silent:
        print(f"Fetching source of {bright_blue}{address}{color} from Etherscan...")

    response = requests.get(
        "https://api.etherscan.io/v2/api", params=params, headers=REQUEST_HEADERS
    )
    if response.status_code != 200:
        raise ConnectionError(
            f"Status {response.status_code} when querying Etherscan: {response.text}"
        )
    data = response.json()
    if int(data["status"]) != 1:
        raise ValueError(f"Failed to retrieve data from API: {data}")

    return data


# console auto-completion logic


def _call_autosuggest(method: ContractCall | ContractTx) -> list[str]:
    # since methods are not unique for each object, we use `__reduce__`
    # to locate the specific object so we can access the correct ABI
    method = method.__reduce__()[1][0]
    return _contract_method_autosuggest(method.abi["inputs"], False, False)


def _transact_autosuggest(method: ContractCall | ContractTx) -> list[str]:
    method = method.__reduce__()[1][0]
    return _contract_method_autosuggest(method.abi["inputs"], True, method.payable)


# assign the autosuggest functionality to various methods
ContractConstructor.encode_input.__dict__["_autosuggest"] = _call_autosuggest
_ContractMethod.call.__dict__["_autosuggest"] = _call_autosuggest
_ContractMethod.encode_input.__dict__["_autosuggest"] = _call_autosuggest

ContractConstructor.estimate_gas.__dict__["_autosuggest"] = _transact_autosuggest
_ContractMethod.estimate_gas.__dict__["_autosuggest"] = _transact_autosuggest
_ContractMethod.transact.__dict__["_autosuggest"] = _transact_autosuggest


def _contract_method_autosuggest(
    args: list[dict[str, Any]], is_transaction: bool, is_payable: bool
) -> list[str]:
    types_list = get_type_strings(args, _fixed168x10)
    names = (i["name"] for i in args)
    suggestions = [f" {typ}{f' {name}' if name else ''}" for name, typ in zip(names, types_list)]
    if is_transaction:
        if is_payable:
            suggestions.append(" {'from': Account", " 'value': Wei}")
        else:
            suggestions.append(" {'from': Account}")
    return suggestions


def _comment_slicer(match: Match) -> str:
    start, mid, end = match.group(1, 2, 3)
    if mid is None:
        # single line comment
        return ""
    elif start is not None or end is not None:
        # multi line comment at start or end of a line
        return ""
    elif "\n" in mid:
        # multi line comment with line break
        return "\n"
    else:
        # multi line comment without line break
        return " "



#!/usr/bin/python3

from collections.abc import Sequence
from typing import Any, Final, cast

from eth_event.main import DecodedEvent, NonDecodedEvent
from eth_typing import ABIComponent, ABIFunction
from faster_eth_abi.grammar import ABIType, TupleType, parse

from brownie.typing import FormattedEvent

from .datatypes import EthAddress, HexString, ReturnValue
from .main import to_bool, to_decimal, to_int, to_string, to_uint
from .utils import get_type_strings

AnyListOrTuple = list[Any] | tuple[Any, ...]

# Internal C constants

_TupleType: Final = TupleType

_parse: Final = parse


def format_input(abi: ABIFunction, inputs: AnyListOrTuple) -> list[Any]:
    """Format contract inputs based on ABI types."""
    abi_inputs = abi["inputs"]
    if len(inputs) and not len(abi_inputs):
        raise TypeError(f"{abi['name']} requires no arguments")
    abi_types = _get_abi_types(abi_inputs)
    try:
        return _format_tuple(abi_types, inputs)
    except Exception as e:
        raise type(e)(f"{abi['name']} {e}") from None


def format_output(abi: ABIFunction, outputs: AnyListOrTuple) -> ReturnValue:
    """Format contract outputs based on ABI types."""
    abi_outputs = abi["outputs"]
    abi_types = _get_abi_types(abi_outputs)
    result = _format_tuple(abi_types, outputs)
    return ReturnValue(result, abi_outputs)


def format_event(event: DecodedEvent | NonDecodedEvent) -> FormattedEvent:
    """Format event data based on ABI types."""
    if not event["decoded"]:
        topics = (
            {"type": "bytes32", "name": name, "value": data}
            for name, data in zip(("topic1", "topic2", "topic3"), event.get("topics", ()))
        )
        event["data"] = [  # type: ignore [typeddict-item]
            *topics,
            {"type": "bytes", "name": "data", "value": _format_single("bytes", event["data"])},
        ]
        event["name"] = "(anonymous)" if "anonymous" in event else "(unknown)"  # type: ignore [typeddict-item]
        return event  # type: ignore [return-value]

    data = event["data"]
    for e in data:
        if not e["decoded"]:
            e["type"] = "bytes32"
            e["name"] += " (indexed)"
    abi_types = _get_abi_types(cast(Sequence[ABIComponent], data))
    event_values = [i["value"] for i in data]
    values = ReturnValue(
        _format_tuple(abi_types, event_values),
        cast(Sequence[ABIComponent], data),
    )
    for e, value in zip(data, values):
        e["value"] = value
    return cast(FormattedEvent, event)


def _format_tuple(abi_types: Sequence[ABIType], values: AnyListOrTuple) -> list[Any]:
    result = []
    _check_array(values, len(abi_types))
    for type_, value in zip(abi_types, values):
        try:
            if type_.is_array:
                result.append(_format_array(type_, value))
            elif isinstance(type_, _TupleType):
                result.append(_format_tuple(type_.components, value))
            else:
                result.append(_format_single(type_.to_type_str(), value))
        except Exception as e:
            raise type(e)(f"'{value}' - {e}") from None
    return result


def _format_array(abi_type: ABIType, values: AnyListOrTuple) -> list[Any]:
    arrlist = cast(tuple[tuple[int, ...], ...], abi_type.arrlist)
    arrlast = arrlist[-1]
    _check_array(values, arrlast[0] if arrlast else None)
    item_type = abi_type.item_type
    if item_type.is_array:
        return [_format_array(item_type, i) for i in values]
    elif isinstance(item_type, _TupleType):
        item_type_components = item_type.components
        return [_format_tuple(item_type_components, i) for i in values]
    item_type_str = item_type.to_type_str()
    return [_format_single(item_type_str, i) for i in values]


def _format_single(type_str: str, value: Any) -> Any:
    # Apply standard formatting to a single value
    if "uint" in type_str:
        return to_uint(value, type_str)
    elif "int" in type_str:
        return to_int(value, type_str)
    elif type_str == "fixed168x10":
        return to_decimal(value)
    elif type_str == "bool":
        return to_bool(value)
    elif type_str == "address":
        return EthAddress(value)
    elif "byte" in type_str:
        return HexString(value, type_str)
    elif "string" in type_str:
        return to_string(value)
    raise TypeError(f"Unknown type: {type_str}")


def _check_array(values: AnyListOrTuple, length: int | None) -> None:
    if not isinstance(values, (list, tuple)):
        # NOTE: we keep this check here in case the user is running in interpreted mode
        raise TypeError(f"Expected list or tuple, got {type(values).__name__}")
    if length is not None and len(values) != length:
        raise ValueError(f"Sequence has incorrect length, expected {length} but got {len(values)}")


def _get_abi_types(abi_params: Sequence[ABIComponent]) -> Sequence[ABIType]:
    if not abi_params:
        return []
    type_str = f"({','.join(get_type_strings(abi_params))})"
    tuple_type = cast(TupleType, _parse(type_str))
    return tuple_type.components



from collections.abc import Callable, Sequence
from typing import Any, final

from web3 import Web3
from web3.types import RPCEndpoint

from brownie._c_constants import regex_findall
from brownie.network.middlewares import BrownieMiddlewareABC


@final
class HardhatMiddleWare(BrownieMiddlewareABC):
    @classmethod
    def get_layer(cls, w3: Web3, network_type: str) -> int | None:
        if w3.client_version.lower().startswith("hardhat"):
            return -100
        else:
            return None

    def process_request(
        self,
        make_request: Callable,
        method: RPCEndpoint,
        params: Sequence[Any],
    ) -> dict[str, Any]:
        result = make_request(method, params)

        # modify Hardhat transaction error to mimic the format that Ganache uses
        if (
            method in {"eth_call", "eth_sendTransaction", "eth_sendRawTransaction"}
            and "error" in result
        ):
            error: dict = result["error"]
            message: str = error["message"]
            if message.startswith("Error: VM Exception") or message.startswith(
                "Error: Transaction reverted"
            ):
                if method == "eth_call":
                    # ganache returns a txid even on a failed eth_call, which is weird,
                    # but we still mimic it here for the sake of consistency
                    txid = "0x"
                else:
                    txid = error["data"]["txHash"]
                data: dict = {}
                error["data"] = {txid: data}
                message = message.split(": ", maxsplit=1)[-1]
                if message == "Transaction reverted without a reason":
                    data.update({"error": "revert", "reason": None})
                elif message.startswith("revert"):
                    data.update({"error": "revert", "reason": message[7:]})
                elif "reverted with reason string '" in message:
                    data.update(error="revert", reason=regex_findall(".*?'(.*)'$", message)[0])
                elif "reverted with an unrecognized custom error" in message:
                    message = message[message.index("0x") : -1]
                    data.update(error="revert", reason=message)
                else:
                    data["error"] = message
        return result



#!/usr/bin/python3
# mypy: disable-error-code="arg-type,union-attr"

import pathlib
import sys
import warnings
from ast import FunctionDef, Import, ImportFrom, dump, parse
from collections.abc import Sequence
from importlib import reload
from importlib.machinery import SourceFileLoader
from importlib.util import find_spec
from pathlib import WindowsPath
from types import FunctionType, ModuleType
from typing import Any, Final

from brownie._c_constants import Path, import_module, sha1
from brownie.exceptions import ProjectNotFound
from brownie.project.main import Project, check_for_project, get_loaded_projects
from brownie.utils import color
from brownie.utils._color import bright_blue, bright_cyan

# C Constants

_FunctionDef: Final = FunctionDef
_Import: Final = Import
_ImportFrom: Final = ImportFrom

_FunctionType: Final = FunctionType

_parse: Final = parse
_dump: Final = dump

_import_module: Final = import_module
_find_spec: Final = find_spec
_reload: Final = reload

_DOT_PATH: Final = Path(".")

_import_cache: Final[dict[str, ModuleType]] = {}


def run(
    script_path: str,
    method_name: str = "main",
    args: Sequence[Any] | None = None,
    kwargs: dict[str, Any] | None = None,
    project: Project | None = None,
    _include_frame: bool = False,
) -> Any:
    """Loads a project script and runs a method in it.

    script_path: path of script to load
    method_name: name of method to run
    args: method args
    kwargs: method kwargs
    project: (deprecated)

    Returns: return value from called method
    """
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}

    script, project = _get_path(script_path)

    # temporarily add project objects to the main namespace, so the script can import them
    if project is not None:
        project._add_to_main_namespace()

    # modify sys.path to ensure script can be imported
    if isinstance(script, WindowsPath):
        # far from elegant but just works
        root_path = str(_DOT_PATH.resolve())
        script = script.relative_to(Path("..").resolve())
    else:
        root_path = _DOT_PATH.resolve().root

    sys.path.insert(0, root_path)

    try:
        module = _import_from_path(script)
        module_name = module.__name__

        func = getattr(module, method_name, None)
        if not isinstance(func, _FunctionType):
            raise AttributeError(f"Module '{module_name}' has no method '{method_name}'")

        file = module.__file__
        absolute_path = Path(file)
        try:
            module_path = absolute_path.relative_to(_DOT_PATH.absolute())
        except ValueError:
            module_path = absolute_path
        print(
            f"\nRunning '{bright_blue}{module_path}{color}::"
            f"{bright_cyan}{method_name}{color}'..."
        )

        if not _include_frame:
            return func(*args, **kwargs)

        # this voodoo preserves the call frame of the function after it has finished executing.
        # we do this so that `brownie run -i` is able to drop into the console with the same
        # namespace as the function that it just ran.

        # first, we extract the source code from the function and parse it to an AST
        # we generate the AST from the entire module to preserve line numbers
        loader = module.__loader__
        assert isinstance(loader, SourceFileLoader)
        source = loader.get_source(module_name) or ""
        func_ast = _parse(source)

        # use the last function with the correct name, consistent with how python handles
        func_nodes: list = [
            i for i in func_ast.body if isinstance(i, _FunctionDef) and i.name == method_name
        ]
        func_ast.body = func_nodes[-1:]

        # next, we insert some new logic into the beginning of the function. this imports
        # the sys module and assigns the current frame to a global var `__brownie_frame`
        injected_source = "import sys\nglobal __brownie_frame\n__brownie_frame = sys._getframe()"
        injected_ast = _parse(injected_source)
        func_ast.body[0].body = injected_ast.body + func_ast.body[0].body  # type: ignore [attr-defined]

        # now we compile the AST into a code object, using the module's `__dict__` as our globals
        # so that we have access to all the required imports and other objects
        f_locals: dict = module.__dict__.copy()
        del f_locals[method_name]
        func_code = compile(func_ast, file, "exec")
        exec(func_code, f_locals)

        # finally, we execute our new function from inside the copied globals dict. the frame
        # is added to the dict as `__global_frame` per our injected code, and we return it for
        # use within the console. so simple!
        return_value = f_locals[method_name](*args, **kwargs)
        return return_value, f_locals["__brownie_frame"]

    finally:
        # cleanup namespace and sys.path
        sys.path.remove(root_path)
        if project is not None:
            project._remove_from_main_namespace()


def _get_path(path_str: str) -> tuple[pathlib.Path, Project | None]:
    # Returns path to a python module
    path = Path(path_str).with_suffix(".py")

    if not get_loaded_projects():
        if not path.exists():
            raise FileNotFoundError(f"Cannot find {path_str}")
        return path.resolve(), None

    if not path.is_absolute():
        for project in get_loaded_projects():
            project_structure: dict = project._structure
            project_scripts = project_structure["scripts"]
            if path.parts[:1] == (project_scripts,):
                script_path = project._path.joinpath(path)
            else:
                script_path = project._path.joinpath(project_scripts).joinpath(path)
            if script_path.exists():
                return script_path.resolve(), project
        raise FileNotFoundError(f"Cannot find {path_str}")

    if not path.exists():
        raise FileNotFoundError(f"Cannot find {path_str}")

    for project in get_loaded_projects():
        if path_str.startswith(project._path.as_posix()):
            return path.resolve(), project

    raise ProjectNotFound(f"{path_str} is not part of an active project")


def _import_from_path(path: pathlib.Path) -> ModuleType:
    # Imports a module from the given path
    import_str = f"{'.'.join(path.parts[1:-1])}.{path.stem}"
    if import_str in _import_cache:
        _reload(_import_cache[import_str])
        return _import_cache[import_str]
    else:
        module = _import_module(import_str)
        _import_cache[import_str] = module
        return module


def _get_ast_hash(path: str | pathlib.Path) -> str:
    # Generates a hash based on the AST of a script.
    with Path(path).open() as fp:
        ast_list = [_parse(fp.read(), path)]
    base_path = str(check_for_project(path))

    for obj in ast_list[0].body:
        if isinstance(obj, _Import):
            name = obj.names[0].name
        elif isinstance(obj, _ImportFrom):
            name = obj.module  # type: ignore [assignment]
        else:
            continue
        try:
            origin = _find_spec(name).origin
        except Exception:
            warnings.warn(
                f"{Path(path).name}, unable to determine import spec for '{name}',"
                " the --update flag may not work correctly with this test file",
                ImportWarning,
            )
            continue
        if origin is not None and base_path in origin:
            with open(origin) as fp:
                ast_list.append(_parse(fp.read(), origin))

    dump = "\n".join(_dump(i) for i in ast_list)
    return sha1(dump.encode()).hexdigest()



from collections.abc import Mapping
from typing import Any, Final, TypeVar, overload

from dotenv.variables import parse_variables

from brownie._c_constants import regex_compile

_T = TypeVar("_T")


def expand_posix_vars(obj: Any, variables: Mapping[str, Any | None]) -> Any:
    """expand_posix_vars recursively expands POSIX values in an object.

    Args:
        obj (any): object in which to interpolate variables.
        variables (dict): dictionary that maps variable names to their value
    """
    if isinstance(obj, (dict,)):
        for key, val in obj.items():
            obj[key] = expand_posix_vars(val, variables)
    elif isinstance(obj, (list,)):
        for index in range(len(obj)):
            obj[index] = expand_posix_vars(obj[index], variables)
    elif isinstance(obj, (str,)):
        obj = _str_to_python_value(_expand(obj, variables))
    return obj


def _expand(value: _T, variables: Mapping = {}) -> _T:
    """_expand does POSIX-style variable expansion

    This is adapted from python-dotenv, specifically here:

    https://github.com/theskumar/python-dotenv/commit/17dba65244c1d4d10f591fe37c924bd2c6fd1cfc

    We need this layer here so we can explicitly pass in variables;
    python-dotenv assumes you want to use os.environ.
    """

    if not isinstance(value, (str,)):
        return value
    atoms = parse_variables(value)
    return "".join(str(atom.resolve(variables)) for atom in atoms)  # type: ignore [return-value]


INT_REGEX: Final = regex_compile(r"^[-+]?[0-9]+$")


@overload
def _str_to_python_value(val: str) -> bool | int | str: ...
@overload
def _str_to_python_value(val: _T) -> _T: ...
def _str_to_python_value(val):
    """_str_to_python_value infers the data type from a string.

    This could eventually use PyYAML's parsing logic.
    """
    if not isinstance(val, (str,)):
        return val
    elif val in {"true", "True", "on"}:
        return True
    elif val in {"false", "False", "off"}:
        return False
    elif INT_REGEX.match(val):
        return int(val)
    return val



"""This file contains utility functions for converting bytes values to hexstrings.

Since this code is frequently accessed, instead of adding runtime checks within
the function bodies, we opted to microoptimize by defining functions specific to
your hexbytes version .
"""

from importlib.metadata import version
from typing import Final

import hexbytes
from eth_typing import HexStr

from brownie._c_constants import HexBytes

HEXBYTES_LT_1_0_0: Final = tuple(int(i) for i in version("hexbytes").split(".")) < (
    1,
    0,
    0,
)
"""
A boolean constant that indicates if the version of your `hexbytes` package is less than 1.0.0.

This exists to address a breaking change in hexbytes v1 and allow brownie to be used with
either version.
"""


def hexbytes_to_hexstring(value: hexbytes.HexBytes) -> HexStr:
    """Convert a HexBytes object to a hex string."""
    # NOTE: this is just a stub for mypy, the func is conditionally
    # defined below based on your hexbytes version


if HEXBYTES_LT_1_0_0:

    def bytes_to_hexstring(value: bytes) -> HexStr:
        """Convert a bytes value to a hexstring on hexbytes<1."""
        return HexBytes(value).hex()

    hexbytes_to_hexstring = HexBytes.hex
    """Convert a HexBytes value to a hexstring on hexbytes<1."""

else:

    def bytes_to_hexstring(value: bytes) -> HexStr:
        """Convert a bytes value to a hexstring on hexbytes>=1."""
        return f"0x{value.hex()}"

    hexbytes_to_hexstring = bytes_to_hexstring
    """Convert a HexBytes value to a hexstring on hexbytes>=1."""



#!/usr/bin/python3

import sys
from subprocess import DEVNULL, PIPE
from typing import Any

import psutil
from requests.exceptions import ConnectionError as RequestsConnectionError

from brownie.exceptions import RPCRequestError
from brownie.network.web3 import web3


def launch(cmd: str, **_: Any) -> None:
    print(f"\nLaunching '{cmd}'...")
    out = DEVNULL if sys.platform == "win32" else PIPE

    return psutil.Popen([cmd], stdin=DEVNULL, stdout=out, stderr=out)


def on_connection() -> None:
    pass


def _request(method: str, args: list) -> int:
    try:
        response = web3.provider.make_request(method, args)
        if "result" in response:
            return response["result"]
    except (AttributeError, RequestsConnectionError):
        raise RPCRequestError("Web3 is not connected.")
    raise RPCRequestError(response["error"]["message"])


def sleep(seconds: int) -> None:
    raise NotImplementedError("Geth dev does not support time travel")


def mine(timestamp: int | None = None) -> None:
    raise NotImplementedError("Geth dev does not support empty mining")


def snapshot() -> int:
    return 0


def revert(snapshot_id: int) -> None:
    raise NotImplementedError("Geth dev does not support snapshots or rewinds")


def unlock_account(address: str) -> None:
    raise NotImplementedError("Geth dev does not support unlocking accounts")



import unittest
import uuid

from brownie._expansion import expand_posix_vars


class TestExpandDict(unittest.TestCase):
    def setUp(self):
        self.v = str(uuid.uuid4())
        self.input = {
            "non": "b",
            "simple": "${FOO}",
            "partial": "the ${FOO}",
            "number": 1,
            "bool": True,
            "nested": {
                "one": "nest ${FOO}",
                "super_nested": {"two": "real ${FOO}", "three": "not"},
            },
            "${A}": "abc",
            "default_envvar_present": "${FOO:-xyz}",
            "default_envvar_missing": "${ABC:-bar}",
            "default_int_present": "${NUM:-42}",
            "default_int_missing": "${ABC:-42}",
            "arr": [{"a": False, "b": False}, {"a": True, "b": "${FOO}"}],
        }
        variables = {"FOO": self.v, "NUM": 314}
        self.res = expand_posix_vars(
            self.input,
            variables,
        )

    def test_basic_string(self):
        assert self.res["non"] == "b"

    def test_simple_expansion(self):
        assert self.res["simple"] == self.v

    def test_partial_string_expansion(self):
        assert self.res["partial"] == f"the {self.v}"

    def test_number(self):
        assert self.res["number"] == 1

    def test_bool(self):
        assert self.res["bool"] == True  # noqa: E712

    def test_nested_partial_string(self):
        assert self.res["nested"]["one"] == f"nest {self.v}"

    def test_double_nested_partial_string(self):
        assert self.res["nested"]["super_nested"]["two"] == f"real {self.v}"

    def test_double_nested_plain(self):
        assert self.res["nested"]["super_nested"]["three"] == "not"

    def test_variable_name_not_expanded(self):
        assert self.res["${A}"] == "abc"

    def test_list_basic(self):
        assert self.res["arr"][0]["a"] == False  # noqa: E712

    def test_list_bool(self):
        assert self.res["arr"][1]["a"] == True  # noqa: E712

    def test_arr_expanded(self):
        assert self.res["arr"][1]["b"] == self.v

    def test_envvar_with_default_value_present(self):
        assert self.res["default_envvar_present"] == self.v

    def test_envvar_with_default_value_missing(self):
        assert self.res["default_envvar_missing"] == "bar"

    def test_envvar_with_default_int_value_present(self):
        assert self.res["default_int_present"] == 314

    def test_envvar_with_default_int_value_missing(self):
        assert self.res["default_int_missing"] == 42
